"""Token and production containers for use in the tokenizer and parser."""

# Can remove once Finesse requires at least Python 3.9.
from __future__ import annotations

from typing import Any, Union
from functools import total_ordering
import re
import abc
from dataclasses import dataclass


@total_ordering
@dataclass(eq=True, frozen=True)
class KatCoordinate:
    """Kat script file coordinate supporting comparison operations."""

    lineno: int
    index: int

    def __lt__(self, other):
        if self.lineno != other.lineno:
            return self.lineno < other.lineno
        return self.index < other.index

    def __add__(self, other):
        return KatCoordinate(self.lineno + other.lineno, self.index + other.index)

    def __sub__(self, other):
        return KatCoordinate(self.lineno - other.lineno, self.index - other.index)

    def __str__(self):
        return f"({self.lineno}, {self.index})"

    def compact(self):
        """Compact coordinate representation.

        Returns
        -------
        :class:`str`
            Compact coordinates.
        """
        return f"{self.lineno}:{self.index}"


@dataclass(eq=True, frozen=True)
class KatBounds:
    """Kat script start and stop bounds.

    This represents a block of text in a kat script file.
    """

    start: KatCoordinate
    stop: KatCoordinate

    def __post_init__(self):
        if self.start > self.stop:
            raise ValueError(f"start ({self.start}) must be < stop ({self.stop})")

    def __str__(self):
        return f"{self.start}:{self.stop}".replace(" ", "")

    def isempty(self):
        """Whether the contents contained within the bounds is empty.

        Returns
        -------
        :class:`bool`
            True if empty, False otherwise.
        """
        return self.start == self.stop

    def lexpand(self, newstart):
        """Expand start boundary.

        Parameters
        ----------
        newstart : :class:`.KatCoordinate`
            The new start coordinate.

        Returns
        -------
        :class:`.KatBounds`
            The expanded bounds.

        Raises
        ------
        ValueError
            If `newstart` is not <= current start coordinate.
        """
        if self.start.lineno < newstart.lineno:
            raise ValueError(
                f"New start {newstart} is not <= current start {self.start}"
            )
        return self.__class__(newstart, self.stop)

    def rexpand(self, newstop):
        """Expand stop boundary.

        Parameters
        ----------
        newstop : :class:`.KatCoordinate`
            The new stop coordinate.

        Returns
        -------
        :class:`.KatBounds`
            The expanded bounds.

        Raises
        ------
        ValueError
            If `newstop` is not >= current stop coordinate.
        """
        if self.stop.lineno > newstop.lineno:
            raise ValueError(f"New stop {newstop} is not >= current stop {self.stop}")
        return self.__class__(self.start, newstop)

    def lcontract(self, newstart):
        """Contract start boundary.

        Parameters
        ----------
        newstart : :class:`.KatCoordinate`
            The new start coordinate.

        Returns
        -------
        :class:`.KatBounds`
            The contracted bounds.

        Raises
        ------
        ValueError
            If `newstart` is not >= current start coordinate.
        """
        if self.start.lineno > newstart.lineno:
            raise ValueError(
                f"New start {newstart} is not >= current start {self.start}"
            )
        return self.__class__(newstart, self.stop)

    def rcontract(self, newstop):
        """Contract stop boundary.

        Parameters
        ----------
        newstop : :class:`.KatCoordinate`
            The new stop coordinate.

        Returns
        -------
        :class:`.KatBounds`
            The contracted bounds.

        Raises
        ------
        ValueError
            If `newstop` is not <= current stop coordinate.
        """
        if self.stop.lineno < newstop.lineno:
            raise ValueError(f"New stop {newstop} is not <= current stop {self.stop}")
        return self.__class__(self.start, newstop)


class Addressable(metaclass=abc.ABCMeta):
    """Mixin defining interface to retrieve strings containing script lines."""

    def script(self, bounds=None):
        """Get the script in the interval [start, stop) defined by `bounds`.

        Lines between `start` and `stop` with no tokens are yielded as empty lines. Gaps
        between the columns spanned by tokens on the same line are yielded as spaces.

        Parameters
        ----------
        bounds : :class:`.KatBounds`, optional
            Bounds within which to retrieve script. Defaults to the whole script.

        Returns
        -------
        :class:`str`
            The script.
        """
        if bounds is None:
            bounds = self.bounds
        return self._script(bounds)

    def script_lines(self, bounds):
        """The script within `bounds` by line.

        Parameters
        ----------
        bounds : :class:`.KatBounds`
            The bounds to retrieve script between.

        Returns
        -------
        :class:`list`
            Lines within `bounds`. Where `bounds` starts or stops mid-way through a
            line, only the part of the line that falls within `bounds` is contained in
            the corresponding line.
        """
        return self.script(bounds).splitlines()

    @abc.abstractmethod
    def _script(self, bounds):
        raise NotImplementedError

    @property
    def start(self):
        """Start coordinate.

        :getter: The start :class:`.KatCoordinate`.
        """
        return self.bounds.start

    @property
    def stop(self):
        """Stop coordinate.

        :getter: The stop :class:`.KatCoordinate`.
        """
        return self.bounds.stop

    @property
    @abc.abstractmethod
    def bounds(self):
        """Container bounds.

        :getter: :class:`.KatBounds` for this container.
        """
        raise NotImplementedError


class TokenContainer(Addressable, metaclass=abc.ABCMeta):
    """Container with concrete token instances."""

    @property
    @abc.abstractmethod
    def tokens(self):
        """Tokens contained in the container.

        :getter: :class:`list` of :class:`.KatToken` objects within this container.
        """
        return []

    @property
    def sorted_tokens(self):
        """Tokens contained in the container, in ascending coordinate order.

        :getter: sorted :class:`list` of :class:`.KatToken` objects within this
                 container.
        """
        return sorted(self.tokens, key=lambda item: item.start)

    @property
    def first_token(self):
        """First container token by coordinate.

        :getter: first :class:`.KatToken` within this container.
        """
        return self.sorted_tokens[0]

    @property
    def last_token(self):
        """Last container token by coordinate.

        :getter: last :class:`.KatToken` within this container.
        """
        return self.sorted_tokens[-1]

    @property
    def bounds(self):
        return KatBounds(self.first_token.start, self.last_token.stop)

    def _script(self, bounds):
        script = ""

        for token in self.sorted_tokens:
            # The code here can't cope with tokens spanning lines. All tokens created by
            # the tokenizer are single-lined, including newline characters (just like
            # Python's tokenizer).
            assert token.start.lineno == token.stop.lineno

            # Tokens are in order, so we can do some simple exit checks.
            if token.stop < bounds.start:
                continue
            if token.start >= bounds.stop:
                break

            value = token.display_value

            if (
                token.lineno == bounds.start.lineno
                and token.start.index < bounds.start.index
            ):
                startindex = bounds.start.index - token.start.index
            else:
                startindex = 0
            if (
                token.lineno == bounds.stop.lineno
                and token.stop.index >= bounds.stop.index
            ):
                stopindex = bounds.stop.index - token.start.index
            else:
                stopindex = None

            script += value[startindex:stopindex]

        return script


@dataclass(eq=True)
class KatFile(Addressable):
    """Container with kat script lines (no concrete tokens)."""

    text: str

    @property
    def _text_lines(self):
        if not self.text:
            # Empty line.
            return [""]
        lines = self.text.splitlines(True)
        if self.text.endswith("\n"):
            # Add a last empty line because splitlines() ignores it, but it affects the
            # file's stop position. See
            # https://docs.python.org/3/library/stdtypes.html#str.splitlines.
            lines.append("")
        return lines

    def _script(self, bounds):
        script = ""
        for lineno, line in enumerate(self._text_lines, start=1):
            if lineno < bounds.start.lineno:
                continue
            if lineno > bounds.stop.lineno:
                break

            if lineno == bounds.start.lineno:
                # Convert column to list index.
                startindex = max(bounds.start.index - 1, 0)
            else:
                startindex = 0
            if lineno == bounds.stop.lineno:
                # Convert column to list index.
                stopindex = max(bounds.stop.index - 1, 0)
            else:
                stopindex = None

            script += line[startindex:stopindex]

        return script

    @property
    def bounds(self):
        start = KatCoordinate(1, 1)
        stop = KatCoordinate(len(self._text_lines), 1 + len(self._text_lines[-1]))
        return KatBounds(start, stop)

    def add(self, string):
        """Add string to end of file.

        Parameters
        ----------
        string : :class:`str`
            The string to add to the end of the file.
        """
        self.text += string


@dataclass(eq=True)
class KatMetaToken(TokenContainer):
    """A token that may not map to a real token.

    Meta tokens are employed by :class:`.KatTokenizer` when converting whitespace and
    newline tokens (provided by :class:`.KatRawTokenizer`) used as delimiters into
    DELIMITER. The use of a single meta token for DELIMITER in this case simplifies the
    parser rules.

    By definition, the value of a meta token does not matter so the class doesn't
    support having one.
    """

    lineno: int
    start_index: int
    stop_index: int
    type: str

    @property
    def tokens(self):
        return super().tokens + [self]

    @property
    def bounds(self):
        return KatBounds(
            KatCoordinate(self.lineno, self.start_index),
            KatCoordinate(self.lineno, self.stop_index),
        )

    def __str__(self):
        return f"__{self.type}__"


@dataclass(eq=True)
class KatToken(KatMetaToken):
    """A real token with the corresponding text's location, type and value."""

    raw_value: Any

    @property
    def value(self):
        return self.raw_value

    @property
    def display_value(self):
        return self.raw_value

    @classmethod
    def with_new_type(cls, tok, new_type):
        """Convert token to a different type.

        Parameters
        ----------
        tok : :class:`.KatToken`
            The token to convert.

        new_type : :class:`str`
            The new token type.

        Returns
        -------
        :class:`.KatToken`
            The kat token, copied from `tok`, with type `new_type`.
        """
        return cls(
            lineno=tok.lineno,
            start_index=tok.start_index,
            stop_index=tok.stop_index,
            type=new_type,
            raw_value=tok.raw_value,
        )

    def __str__(self):
        return repr(self.display_value)


@dataclass(eq=True)
class BaseCustomToken(KatToken, metaclass=abc.ABCMeta):
    """A custom token."""

    def __post_init__(self):
        # Ensure the value is valid. Ensures errors in determining the token's value get
        # thrown on instantiation rather than when later read.
        try:
            self.value
        except ValueError:
            # NOTE: the actual syntax that's incorrect will be added by the tokenizer
            # when it catches this.
            raise SyntaxError("syntax error")


@dataclass(eq=True)
class KatWhitespaceToken(BaseCustomToken):
    """A whitespace token."""

    # Tab size, used for displaying parsed script.
    TAB_SIZE = 4

    @property
    def display_value(self):
        # Convert tabs into the same number of spaces assumed by the tokenizer. This
        # avoids issues when displaying script inside error messaages.
        return self.raw_value.replace("\t", "~" * self.TAB_SIZE)


@dataclass(eq=True)
class KatBooleanToken(BaseCustomToken):
    """A boolean token."""

    @property
    def value(self):
        return self.raw_value == "True" or self.raw_value == "true"


@dataclass(eq=True)
class KatReferenceToken(BaseCustomToken):
    """A reference token."""

    @property
    def value(self):
        return self.raw_value.lstrip("&")


@dataclass(eq=True)
class KatNumberToken(BaseCustomToken):
    """A number token."""

    SI_NUMBER_PATTERN = re.compile(r".*[pnumkMGT]$")

    @property
    def value(self):
        value = self.raw_value

        if "j" in value:
            return complex(value)

        if self.SI_NUMBER_PATTERN.match(value):
            value = value.replace("p", "e-12")
            value = value.replace("n", "e-9")
            value = value.replace("u", "e-6")
            value = value.replace("m", "e-3")
            value = value.replace("k", "e3")
            value = value.replace("M", "e6")
            value = value.replace("G", "e9")
            value = value.replace("T", "e12")

        float_value = float(value)

        if "e" not in value.lower() and float_value.is_integer():
            # Only convert real integers
            value = value.rstrip(".")
            if value.endswith(".0"):
                value = value[:-2]
            return int(value)

        return float_value


@dataclass(eq=True)
class KatStringToken(BaseCustomToken):
    """A string token."""

    @property
    def value(self):
        value = self.raw_value

        # Get rid of escape characters.
        value = value.replace("\\n", "\n")
        value = value.replace("\\r", "\r")
        value = value.replace("\\", "")
        # Get rid of the quotes.
        value = value[1:-1]

        return value


@dataclass(eq=True)
class KatNoneToken(BaseCustomToken):
    """A null token."""

    @property
    def value(self):
        return None


@dataclass(eq=True)
class ArgumentContainer(TokenContainer, metaclass=abc.ABCMeta):
    """Mixin for containers that contain arguments."""

    arguments: list[TokenContainer]

    @property
    def tokens(self):
        tokens = super().tokens
        for argument in self.arguments:
            tokens.extend(argument.tokens)
        return tokens


@dataclass(eq=True)
class ExtraTokenContainer(TokenContainer, metaclass=abc.ABCMeta):
    """Mixin for containers that contain extra tokens."""

    extra: list[KatToken]

    @property
    def tokens(self):
        return super().tokens + self.extra


@dataclass(eq=True)
class KatScript(ArgumentContainer, ExtraTokenContainer, TokenContainer):
    """Represents a kat script."""


@dataclass(eq=True)
class KatScriptItem(TokenContainer, metaclass=abc.ABCMeta):
    """Represents a top level kat script item."""

    directive: KatToken

    @property
    @abc.abstractmethod
    def node_name(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def unique_name_token(self):
        raise NotImplementedError

    @property
    def tokens(self):
        return super().tokens + [self.directive]


@dataclass(eq=True)
class KatElement(ArgumentContainer, ExtraTokenContainer, KatScriptItem):
    """Represents a parsed element statement and any corresponding arguments."""

    name: KatToken

    @property
    def tokens(self):
        return super().tokens + [self.name]

    @property
    def node_name(self):
        return self.name.display_value

    @property
    def unique_name_token(self):
        return self.name

    def __str__(self):
        name = f"{self.directive.display_value} {self.name.display_value}"
        args = f"{', '.join([str(arg) for arg in self.arguments])}"
        return f"{name} ( {args} )"


@dataclass(eq=True)
class KatFunction(ArgumentContainer, ExtraTokenContainer, KatScriptItem):
    """Represents a parsed kat function statement and any corresponding arguments."""

    @property
    def name(self):
        """The token that represents the function name.

        This is used by the error handler for invalid keyword arguments, for example.
        """
        return self.directive

    @property
    def node_name(self):
        return f"{self.directive.display_value}[{self.directive.start.compact()}]"

    @property
    def unique_name_token(self):
        return self.directive

    def __str__(self):
        return f"{self.directive}({', '.join([str(argument) for argument in self.arguments])})"


@dataclass(eq=True)
class KatKwarg(ExtraTokenContainer, TokenContainer):
    """Represents a kat argument containing a key, value and '='."""

    key: KatToken
    value: Union[KatToken, KatFunction, KatExpression, KatGroupedExpression, KatArray]

    @property
    def tokens(self):
        tokens = super().tokens
        if self.key:
            tokens.append(self.key)
        tokens.extend(self.value.tokens)
        return tokens

    @property
    def is_positional(self):
        return self.key is None

    @property
    def is_keyword(self):
        return self.key is not None

    def __str__(self):
        if self.is_keyword:
            return f"{self.key}={self.value}"
        return f"{self.value}"


@dataclass(eq=True)
class KatExpression(ArgumentContainer, ExtraTokenContainer, TokenContainer):
    """Represents a kat script expression."""

    operator: KatToken

    @property
    def tokens(self):
        return super().tokens + [self.operator]

    @property
    def lhs(self):
        return self.arguments[0]

    @property
    def rhs(self):
        return self.arguments[1]

    def __str__(self):
        return f"{self.lhs}{self.operator}{self.rhs}"


@dataclass(eq=True)
class KatGroupedExpression(ArgumentContainer, ExtraTokenContainer, TokenContainer):
    """Represents a kat script expression group, i.e. `(<expression>)`."""

    @property
    def expression(self):
        return self.arguments[0]

    def __str__(self):
        return f"({self.expression})"


@dataclass(eq=True)
class KatArray(ArgumentContainer, ExtraTokenContainer, TokenContainer):
    """Represents a kat script array."""

    def __str__(self):
        return f"[{', '.join([str(argument) for argument in self.arguments])}]"


@dataclass(eq=True)
class KatNumericalArray(KatArray):
    """Represents a kat script numerical array."""

    def __str__(self):
        return f"ndarray([{', '.join([str(argument) for argument in self.arguments])}])"

    @classmethod
    def from_array(cls, array):
        """Convert array to numerical array."""
        return cls(arguments=array.arguments, extra=array.extra)



"""Detectors for calculating mode mismatches in an optical configuration."""

import numpy as np

from ..components.node import OpticalNode, Port
from .general import Detector
from .compute.gaussian import ModeMismatchDetectorWorkspace


class ModeMismatchDetector(Detector):
    """Detector for mode mismatch figure-of-merit for a specified node coupling.

    The computed quantity is given in :meth:`finesse.gaussian.BeamParam.mismatch` where
    :math:`q_1` is the input beam parameter at `node1` propagated via the associated
    ABCD matrix to `node2`, and :math:`q_2` is the beam parameter at `node2`.

    As mode mismatches cannot occur over spaces in Finesse, `node1` must be an input
    node and `node2` must be an output node.

    Parameters
    ----------
    name : str
        Name of the detector.

    node1 : :class:`.OpticalNode` or :class:`.Port`
        Input node or port. If a :class:`.Port` instance is given then this
        node will be the input node of that port. Note that this node cannot
        be an output node.

    node2 : :class:`.OpticalNode` or :class:`.Port`
        Output node or port. If a :class:`.Port` instance is given then this
        node will be the output node of that port. Note that this node cannot
        be an input node.

    direction : str, optional; default: "x"
        Plane of computation, defaults to "x" for the tangential plane. Changing
        to "y" will compute the mode mismatch in the sagittal plane.

    percent : bool, optional; default: True
        Whether to calculate the mode mismatch as a fraction (default behaviour)
        or a percentage.
    """

    def __init__(self, name, node1, node2, direction="x", percent=False):
        if isinstance(node1, Port):
            node1 = node1.i
        if isinstance(node2, Port):
            node2 = node2.o

        if not isinstance(node1, OpticalNode):
            raise TypeError("Argument node1 must be an OpticalNode.")

        if not isinstance(node2, OpticalNode):
            raise TypeError("Argument node2 must be an OpticalNode.")

        if not node1.is_input:
            raise ValueError("Expected node1 to be an input node.")
        if node2.is_input:
            raise ValueError("Expected node2 to be an output node.")

        if node1.component != node2.component:
            raise ValueError(
                "Expected node1 and node2 to be owned by the same component."
            )

        label = f"Mode mismatch\n{node1.full_name}->{node2.full_name}"
        unit = "%" if percent else ""
        super().__init__(name, dtype=np.float64, label=label, unit=unit)

        comp = node1.component
        comp.check_coupling(node1, node2)

        self.__node1 = node1
        self.__node2 = node2
        self.__direction = direction
        self.__percent = percent

    @property
    def in_node(self):
        """Input node.

        :getter: Returns the input node. Read-only.
        """
        return self.__node1

    @property
    def out_node(self):
        """Output node.

        :getter: Returns the output node. Read-only.
        """
        return self.__node2

    @property
    def component(self):
        """The component at which the mode mismatch is calculated.

        :getter: Returns the associated component. Read-only.
        """
        return self.in_node.component

    @property
    def in_percent(self):
        """Flag indicating whether the mismatch is computed as a percentage.

        :getter: Returns the percentage flag.
        :setter: Sets the percentage flag.
        """
        return self.__percent

    @in_percent.setter
    def in_percent(self, value):
        self.__percent = bool(value)

    @property
    def direction(self):
        """The plane in which the mode mismatch is calculated.

        A value of "x" represents the tangential plane, whilst "y" gives
        the sagittal plane.

        :getter: Returns the plane of computation as a string representation. Read-only.
        """
        return self.__direction

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return True

    def _get_workspace(self, sim):
        return ModeMismatchDetectorWorkspace(self, sim)



from finesse import Model
import pytest


@pytest.fixture()
def model():
    """Fixture with module scope for quick fuzzing tests."""
    return Model()


def test_remove_mirror(model):
    model.parse("mirror m1")
    assert "m1" in model.elements
    model.remove(model.m1)
    assert "m1" not in model.elements


@pytest.mark.xfail
def test_remove_space(model):
    model.parse("mirror m1")
    model.parse("mirror m2")
    model.parse("space s1 portA=m1.p2 portB=m2.p1")
    assert "s1" in model.elements
    model.remove(model.elements["s1"])
    assert "s1" not in model.elements


def test_string(model):
    model.parse("mirror m1")
    assert "m1" in model.elements
    model.remove("m1")
    assert "m1" not in model.elements


def test_error_on_unrecognised(model):
    with pytest.raises(TypeError, match=r".*not recongised.*"):
        model.remove(10)



import pytest
from finesse.script.containers import (
    KatScript,
    KatElement,
    KatFunction,
    KatExpression,
    KatGroupedExpression,
    KatNumericalArray,
)
from ..util import (
    NAME,
    NUMBER,
    SPACE,
    COMMA,
    PLUS,
    MINUS,
    TIMES,
    DIVIDE,
    FLOORDIVIDE,
    POWER,
    LPAREN,
    RPAREN,
    LBRACKET,
    RBRACKET,
)


@pytest.mark.parametrize(
    "expression,expected",
    (
        (
            "1+2",
            KatExpression(
                operator=PLUS(1, 19),
                arguments=[NUMBER(1, 18, "1"), NUMBER(1, 20, "2")],
                extra=[],
            ),
        ),
        (
            "3.6-2.2",
            KatExpression(
                operator=MINUS(1, 21),
                arguments=[NUMBER(1, 18, "3.6"), NUMBER(1, 22, "2.2")],
                extra=[],
            ),
        ),
        (
            "3.141*10",
            KatExpression(
                operator=TIMES(1, 23),
                arguments=[NUMBER(1, 18, "3.141"), NUMBER(1, 24, "10")],
                extra=[],
            ),
        ),
        (
            "5.04/1",
            KatExpression(
                operator=DIVIDE(1, 22),
                arguments=[NUMBER(1, 18, "5.04"), NUMBER(1, 23, "1")],
                extra=[],
            ),
        ),
        (
            "4//8.",
            KatExpression(
                operator=FLOORDIVIDE(1, 19),
                arguments=[NUMBER(1, 18, "4"), NUMBER(1, 21, "8.")],
                extra=[],
            ),
        ),
        (
            "8.4**3.0",
            KatExpression(
                operator=POWER(1, 21),
                arguments=[NUMBER(1, 18, "8.4"), NUMBER(1, 23, "3.0")],
                extra=[],
            ),
        ),
    ),
)
def test_binary_expression(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,expected",
    (
        (
            "1+2+3",
            KatExpression(
                operator=PLUS(1, 21),
                arguments=[
                    KatExpression(
                        operator=PLUS(1, 19),
                        arguments=[NUMBER(1, 18, "1"), NUMBER(1, 20, "2")],
                        extra=[],
                    ),
                    NUMBER(1, 22, "3"),
                ],
                extra=[],
            ),
        ),
        (
            "-1+2-3+4",
            KatExpression(
                operator=PLUS(1, 24),
                arguments=[
                    KatExpression(
                        operator=MINUS(1, 22),
                        arguments=[
                            KatExpression(
                                operator=PLUS(1, 20),
                                arguments=[
                                    KatFunction(
                                        directive=MINUS(1, 18),
                                        arguments=[NUMBER(1, 19, "1")],
                                        extra=[],
                                    ),
                                    NUMBER(1, 21, "2"),
                                ],
                                extra=[],
                            ),
                            NUMBER(1, 23, "3"),
                        ],
                        extra=[],
                    ),
                    NUMBER(1, 25, "4"),
                ],
                extra=[],
            ),
        ),
    ),
)
def test_nested_expression(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,expected",
    (
        (
            "(1)",
            KatGroupedExpression(
                arguments=[NUMBER(1, 19, "1")], extra=[LPAREN(1, 18), RPAREN(1, 20)]
            ),
        ),
        (
            "(((1)))",
            KatGroupedExpression(
                arguments=[
                    KatGroupedExpression(
                        arguments=[
                            KatGroupedExpression(
                                arguments=[NUMBER(1, 21, "1")],
                                extra=[LPAREN(1, 20), RPAREN(1, 22)],
                            ),
                        ],
                        extra=[LPAREN(1, 19), RPAREN(1, 23)],
                    ),
                ],
                extra=[LPAREN(1, 18), RPAREN(1, 24)],
            ),
        ),
        (
            "(((((1)))))",
            KatGroupedExpression(
                arguments=[
                    KatGroupedExpression(
                        arguments=[
                            KatGroupedExpression(
                                arguments=[
                                    KatGroupedExpression(
                                        arguments=[
                                            KatGroupedExpression(
                                                arguments=[NUMBER(1, 23, "1")],
                                                extra=[LPAREN(1, 22), RPAREN(1, 24)],
                                            ),
                                        ],
                                        extra=[LPAREN(1, 21), RPAREN(1, 25)],
                                    ),
                                ],
                                extra=[LPAREN(1, 20), RPAREN(1, 26)],
                            ),
                        ],
                        extra=[LPAREN(1, 19), RPAREN(1, 27)],
                    ),
                ],
                extra=[LPAREN(1, 18), RPAREN(1, 28)],
            ),
        ),
    ),
)
def test_parentheses(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,expected",
    (
        # Standard precendence.
        (
            "2+3*4",
            KatExpression(
                operator=PLUS(1, 19),
                arguments=[
                    NUMBER(1, 18, "2"),
                    KatExpression(
                        operator=TIMES(1, 21),
                        arguments=[NUMBER(1, 20, "3"), NUMBER(1, 22, "4")],
                        extra=[],
                    ),
                ],
                extra=[],
            ),
        ),
        (
            "3*4+2",
            KatExpression(
                operator=PLUS(1, 21),
                arguments=[
                    KatExpression(
                        operator=TIMES(1, 19),
                        arguments=[NUMBER(1, 18, "3"), NUMBER(1, 20, "4")],
                        extra=[],
                    ),
                    NUMBER(1, 22, "2"),
                ],
                extra=[],
            ),
        ),
        (
            "3+5**2",
            KatExpression(
                operator=PLUS(1, 19),
                arguments=[
                    NUMBER(1, 18, "3"),
                    KatExpression(
                        operator=POWER(1, 21),
                        arguments=[NUMBER(1, 20, "5"), NUMBER(1, 23, "2"),],
                        extra=[],
                    ),
                ],
                extra=[],
            ),
        ),
        (
            "4**3**2",
            KatExpression(
                operator=POWER(1, 19),
                arguments=[
                    NUMBER(1, 18, "4"),
                    KatExpression(
                        operator=POWER(1, 22),
                        arguments=[NUMBER(1, 21, "3"), NUMBER(1, 24, "2")],
                        extra=[],
                    ),
                ],
                extra=[],
            ),
        ),
        (
            "10/5/2",
            KatExpression(
                operator=DIVIDE(1, 22),
                arguments=[
                    KatExpression(
                        operator=DIVIDE(1, 20),
                        arguments=[NUMBER(1, 18, "10"), NUMBER(1, 21, "5")],
                        extra=[],
                    ),
                    NUMBER(1, 23, "2"),
                ],
                extra=[],
            ),
        ),
        # Bracket-enforced precendence.
        (
            "(3*5)**2",
            KatExpression(
                operator=POWER(1, 23),
                arguments=[
                    KatGroupedExpression(
                        arguments=[
                            KatExpression(
                                operator=TIMES(1, 20),
                                arguments=[NUMBER(1, 19, "3"), NUMBER(1, 21, "5")],
                                extra=[],
                            ),
                        ],
                        extra=[LPAREN(1, 18), RPAREN(1, 22)],
                    ),
                    NUMBER(1, 25, "2"),
                ],
                extra=[],
            ),
        ),
        (
            "(2+3)*4",
            KatExpression(
                operator=TIMES(1, 23),
                arguments=[
                    KatGroupedExpression(
                        arguments=[
                            KatExpression(
                                operator=PLUS(1, 20),
                                arguments=[NUMBER(1, 19, "2"), NUMBER(1, 21, "3")],
                                extra=[],
                            ),
                        ],
                        extra=[LPAREN(1, 18), RPAREN(1, 22)],
                    ),
                    NUMBER(1, 24, "4"),
                ],
                extra=[],
            ),
        ),
        (
            "(4**3)**2",
            KatExpression(
                operator=POWER(1, 24),
                arguments=[
                    KatGroupedExpression(
                        arguments=[
                            KatExpression(
                                operator=POWER(1, 20),
                                arguments=[NUMBER(1, 19, "4"), NUMBER(1, 22, "3")],
                                extra=[],
                            ),
                        ],
                        extra=[LPAREN(1, 18), RPAREN(1, 23)],
                    ),
                    NUMBER(1, 26, "2"),
                ],
                extra=[],
            ),
        ),
        (
            "10/(5/2)",
            KatExpression(
                operator=DIVIDE(1, 20),
                arguments=[
                    NUMBER(1, 18, "10"),
                    KatGroupedExpression(
                        arguments=[
                            KatExpression(
                                operator=DIVIDE(1, 23),
                                arguments=[NUMBER(1, 22, "5"), NUMBER(1, 24, "2")],
                                extra=[],
                            ),
                        ],
                        extra=[LPAREN(1, 21), RPAREN(1, 25)],
                    ),
                ],
                extra=[],
            ),
        ),
    ),
)
def test_precendence(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "sign,expression",
    (
        # Integers.
        ("+", "0"),
        ("-", "0"),
        ("+", "15"),
        ("-", "15"),
        ("+", "50505"),
        ("-", "50505"),
        ## Long.
        ("+", "123141242151251616110"),
        ("-", "123141242151251616110"),
        # Floats.
        ("+", "0.0"),
        ("-", "0.0"),
        ("+", "15.151"),
        ("-", "15.151"),
        ## Scientific.
        ("+", "0.e7"),
        ("-", "0.e7"),
        ("+", "0.0e7"),
        ("-", "0.0e7"),
        ("+", "1.23e7"),
        ("-", "1.23e7"),
        ("+", "1.23e+7"),
        ("-", "1.23e+7"),
        ("+", "1.23e-7"),
        ("-", "1.23e-7"),
        ## Infinities.
        ("+", "inf"),
        ("-", "inf"),
        ## Imaginaries.
        ("+", "0j"),
        ("-", "0j"),
        ("+", "0.j"),
        ("-", "0.j"),
        ("+", "0.0j"),
        ("-", "0.0j"),
        ("+", "10j"),
        ("-", "10j"),
        ("+", "1.32j"),
        ("-", "1.32j"),
        ## Scientific.
        ("+", "0e7j"),
        ("-", "0e7j"),
        ("+", "0.e7j"),
        ("-", "0.e7j"),
        ("+", "0.0e7j"),
        ("-", "0.0e7j"),
        ("+", "1.23e7j"),
        ("-", "1.23e7j"),
        ("+", "1.23e+7j"),
        ("-", "1.23e+7j"),
        ("+", "1.23e-7j"),
        ("-", "1.23e-7j"),
        ## Infinities.
        ("+", "infj"),
        ("-", "infj"),
    ),
)
def test_unary_number(parser, sign, expression):
    """Test unary numbers.

    Unary numbers like -1 get parsed into the function MINUS(1). That means the number is always
    parsed without the sign and the sign determines the function (MINUS or PLUS).
    """
    assert parser.parse(f"fake_element el1 {sign}{expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatFunction(
                        directive=MINUS(1, 18) if sign == "-" else PLUS(1, 18),
                        arguments=[NUMBER(1, 19, expression)],
                        extra=[],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,expected",
    (
        (
            # Empty functions parse instead as statements (via the `nestable_statement` production).
            # This is corrected in the build step.
            "a()",
            KatFunction(
                directive=NAME(1, 18, "a"),
                arguments=[],
                extra=[LPAREN(1, 19), RPAREN(1, 20)],
            ),
        ),
        (
            "a(3.141)",
            KatFunction(
                directive=NAME(1, 18, "a"),
                arguments=[NUMBER(1, 20, "3.141")],
                extra=[LPAREN(1, 19), RPAREN(1, 25)],
            ),
        ),
        (
            "b(1, 2, 3)",
            KatFunction(
                directive=NAME(1, 18, "b"),
                arguments=[NUMBER(1, 20, "1"), NUMBER(1, 23, "2"), NUMBER(1, 26, "3"),],
                extra=[
                    LPAREN(1, 19),
                    COMMA(1, 21),
                    SPACE(1, 22),
                    COMMA(1, 24),
                    SPACE(1, 25),
                    RPAREN(1, 27),
                ],
            ),
        ),
        (
            "a(a(a(1), 2), 3, 4)",
            KatFunction(
                directive=NAME(1, 18, "a"),
                arguments=[
                    KatFunction(
                        directive=NAME(1, 20, "a"),
                        arguments=[
                            KatFunction(
                                directive=NAME(1, 22, "a"),
                                arguments=[NUMBER(1, 24, "1")],
                                extra=[LPAREN(1, 23), RPAREN(1, 25)],
                            ),
                            NUMBER(1, 28, "2"),
                        ],
                        extra=[
                            LPAREN(1, 21),
                            COMMA(1, 26),
                            SPACE(1, 27),
                            RPAREN(1, 29),
                        ],
                    ),
                    NUMBER(1, 32, "3"),
                    NUMBER(1, 35, "4"),
                ],
                extra=[
                    LPAREN(1, 19),
                    COMMA(1, 30),
                    SPACE(1, 31),
                    COMMA(1, 33),
                    SPACE(1, 34),
                    RPAREN(1, 36),
                ],
            ),
        ),
    ),
)
def test_function(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,expected",
    (
        (
            "2*[1, 2]",
            KatExpression(
                operator=TIMES(1, 19),
                arguments=[
                    NUMBER(1, 18, "2"),
                    KatNumericalArray(
                        arguments=[NUMBER(1, 21, "1"), NUMBER(1, 24, "2")],
                        extra=[
                            LBRACKET(1, 20),
                            COMMA(1, 22),
                            SPACE(1, 23),
                            RBRACKET(1, 25),
                        ],
                    ),
                ],
                extra=[],
            ),
        ),
        (
            "2*[1, 2/[3]]",
            KatExpression(
                operator=TIMES(1, 19),
                arguments=[
                    NUMBER(1, 18, "2"),
                    KatNumericalArray(
                        arguments=[
                            NUMBER(1, 21, "1"),
                            KatExpression(
                                operator=DIVIDE(1, 25),
                                arguments=[
                                    NUMBER(1, 24, "2"),
                                    KatNumericalArray(
                                        arguments=[NUMBER(1, 27, "3")],
                                        extra=[LBRACKET(1, 26), RBRACKET(1, 28)],
                                    ),
                                ],
                                extra=[],
                            ),
                        ],
                        extra=[
                            LBRACKET(1, 20),
                            COMMA(1, 22),
                            SPACE(1, 23),
                            RBRACKET(1, 29),
                        ],
                    ),
                ],
                extra=[],
            ),
        ),
    ),
)
def test_numerical_array(parser, expression, expected):
    assert parser.parse(f"fake_element el1 {expression}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[expected],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )



"""CLI run command tests."""

from tempfile import NamedTemporaryFile
from finesse.__main__ import run
from finesse.script import parse_file
from .util import sanitized_output


def test_run_same_as_python_api(isolated_cli, input_file):
    """Test that running kat script via CLI same as calling :meth:`finesse.script.parse_file`."""
    cli_result = isolated_cli.invoke(run, [input_file, "--no-plot"])

    api_model = parse_file(input_file)
    api_solution = api_model.run()

    assert str(api_solution) in sanitized_output(cli_result)
    assert cli_result.exit_code == 0


def test_trace_flag(isolated_cli, input_file):
    """Test the --trace run flag."""
    cli_result = isolated_cli.invoke(run, [input_file, "--no-plot", "--trace"])
    assert "Trace:" in sanitized_output(cli_result)
    assert cli_result.exit_code == 0


def test_no_trace_flag(isolated_cli, input_file):
    """Test the --trace run flag."""
    cli_result = isolated_cli.invoke(run, [input_file, "--no-plot", "--no-trace"])
    assert "Trace:" not in sanitized_output(cli_result)
    assert cli_result.exit_code == 0


def test_parse_error(isolated_cli, tmp_path):
    """Test that parsing errors appear in the output."""
    tmp_file = str(tmp_path / "kat.tmp")
    with open(tmp_file, "w") as fobj:
        fobj.write("laser l1 __fake_param__=1")
    cli_result = isolated_cli.invoke(run, [tmp_file, "--no-plot"])
    assert "ERROR:" in sanitized_output(cli_result)
    assert cli_result.exit_code > 0



"""Actions."""

import abc
import re
import logging
import textwrap
from collections import defaultdict
import numpy as np
from tabulate import tabulate

import finesse
from finesse.solutions import ArraySolution
from finesse.parameter import Parameter
from finesse.analysis.runners import run_axes_scan, run_fsig_sweep
from finesse.tree import TreeNode
from finesse.solutions import BaseSolution
from finesse.parameter import GeometricParameter
from finesse.element import ModelElement
from finesse.components import Port, Node, NodeType, DegreeOfFreedom
from finesse.detectors.compute.quantum import QShot0Workspace, QShotNWorkspace

LOGGER = logging.getLogger(__name__)


def convert_str_to_parameter(model, attr: str):
    """Converts names `component.parameter` or `component` to a parameter object. Will
    return default parameter when component name is given.

    Parameters
    ----------
    model : Model
        Model object to look for parameter in
    attr : str
        String value for the name of an element or a parameters full name

    Returns
    -------
    parameter
        The equivalent Parameter object for the attr provided
    """
    obj = model.reduce_get_attr(attr)
    # If this attr string has no period in it, assume it is an element name
    # and try and get it
    if "." in attr:
        return obj
    else:
        if obj.default_parameter_name is None:
            raise ValueError(
                f"{repr(obj)} does not have a default parameter, please specify one to use"
            )
        return getattr(obj, obj.default_parameter_name)


def get_sweep_array(start: float, stop: float, steps: int, mode="lin"):
    start = float(start)
    stop = float(stop)
    steps = int(steps)
    if steps <= 0:
        raise Exception("Steps must be greater than 0")

    if mode == "lin":
        arr = np.linspace(start, stop, steps + 1)
    else:
        arr = np.logspace(np.log10(start), np.log10(stop), steps + 1)

    return arr


def request_dict_reduction(A, B):
    dd = defaultdict(list)
    for d in (A, B):
        for key, value in d.items():
            dd[key].extend(value)
    return dd


class AnalysisState(TreeNode):
    def __init__(self, model, name="AnalysisState", parent=None):
        super().__init__(f"{name} {model}", parent=parent)
        assert isinstance(model, finesse.model.Model)
        self.__model = model
        self.__sim = None
        self.__previous_solution = None
        self.model_finished_with = True

    @property
    def model(self):
        return self.__model

    @property
    def sim(self):
        return self.__sim

    @property
    def previous_solution(self):
        return self.__previous_solution

    def apply(self, action):
        sol = action._do(self)
        if sol:
            self.__previous_solution = sol
        return sol

    def _split(self):
        state = AnalysisState(self.model.deepcopy(), parent=self)
        return state

    def name_to_elec_mech_nodes(self, nodes):
        rtn = []
        for name in nodes:
            obj = self.model.reduce_get_attr(name)
            is_port = isinstance(obj, Port)
            is_node = isinstance(obj, Node)
            is_dof = isinstance(obj, DegreeOfFreedom)

            if is_dof:
                obj = obj.AC
                is_port = True
            elif is_port and len(obj.nodes) != 1:
                raise ValueError(
                    f"Port {repr(obj)} does not have a single node ({obj.nodes}) so you must specify which to use."
                )
            elif not (is_node or is_port):
                raise ValueError(f"Value {repr(obj)} was neither a Port nor a Node")
            elif obj.type == NodeType.OPTICAL:
                raise ValueError(
                    f"Optical nodes/ports {obj} cannot be used for excitations or outputs"
                )

            if is_port:  # select only single node from port
                obj = obj.nodes[0]
            obj.used_in_detector_output.append(self)
            rtn.append(obj)
        return rtn

    def _build_model(self, changing_params, keep_nodes):
        if not self.model_finished_with:
            raise Exception(
                "Trying to build new model whilst current one is in use. Make sure to call `finished()` on this state if the simulation has been completed."
            )

        if self.model.is_built:
            self._finished()

        LOGGER.info(
            f"Building simulation for model {repr(self.model)}"
            f"Changing parameters = {changing_params}"
        )

        # If we do not have a simulation we need to build one
        for p in changing_params:
            p.is_tunable = True

        self.keep_nodes = tuple(self.name_to_elec_mech_nodes(keep_nodes))

        self.__changing_params = changing_params
        self.__sim = self.model._build()
        self.__sim.__enter__()
        self.model_finished_with = False

    def _finished(self):
        if self.__sim:
            LOGGER.info(
                f"Finishing simulation {repr(self.sim)} for model {repr(self.model)}"
            )
            self.model_finished_with = True
            self.__sim.__exit__(None, None, None)
            self.model.unbuild()
            for p in self.__changing_params:
                p.is_tunable = False
            for obj in self.keep_nodes:
                obj.used_in_detector_output.remove(self)
            self.__sim = None

    def __copy__(self):
        raise Exception("Cannot copy state objects")

    def __deepcopy__(self):
        raise Exception("Cannot copy state objects")


class Action(metaclass=abc.ABCMeta):
    def __init__(self, name, analysis_state_manager=False):
        self.__name = name
        self.__analysis_state_manager = analysis_state_manager

    @property
    def name(self):
        return self.__name

    @property
    def analysis_state_manager(self):
        return self.__analysis_state_manager

    def run(self, model, return_state=False):
        """
        Parameters
        ----------
        model : Model
            Model to run this action on
        return_state : boolean
            If True the AnalysisState object is returned along with the solution

        Returns
        -------
        solution : BaseSolution
            Solution object generated by this action
        state : AnalysisState, when return_state = True
            The final state object after pasing through the action. This can be used
            to extract the models generated and tuned at later actions.
        """
        state = AnalysisState(model)
        try:
            if not self.analysis_state_manager:
                action = Series(self)
            else:
                action = self

            result = state.apply(action)

            if type(result) is tuple:
                sol = BaseSolution("root")
                for _ in result:
                    if _ is not None:
                        sol.add(_)
            else:
                sol = result

            if type(sol) is BaseSolution and len(sol.children) == 1:
                sol = sol[0]
        finally:
            state._finished()

        if return_state:
            return sol, state
        else:
            return sol

    @abc.abstractmethod
    def _requests(self, model, memo, first=True):
        """Updates the memo dictionary with details about what this action needs from a
        simulation to run. Parent actions will get requests from all its child actions
        so that it can build a model that suits all of them, to minimise the amount of
        building.

        This method can do initial checks to make sure the model has the
        required features to perform the action too.

        memo['changing_parameters'] - append to this list the full name string
                                      of parameters that this action needs
        memo['keep_nodes'] - append to this list the full name string
                                    of nodes that this action needs to keep.
                                    This should be used where actions are
                                    accessing node outputs without using a
                                    detector element (which registers that
                                    nodes should be kept already).

        Parameters
        ----------
        model : Model
            The Model that the action will be operating on
        memo : defaultdict(list)
            A dictionary that should be filled with requests
        first : boolean
            True if this is the first request being made
        """
        raise NotImplementedError()

    def get_requests(self, model):
        memo = defaultdict(list)
        self._requests(model, memo)
        return memo

    @abc.abstractmethod
    def _do(self, state: AnalysisState) -> BaseSolution:
        pass

    def plan(self, previous=None):
        """Returns an expected plan for the actions that will be run in a tree form.
        This may not be exactly what is ran.

        Returns
        -------
        plan : TreeNode
        """
        if previous is None:
            previous = TreeNode("start")

        me = TreeNode(self.name)
        me.empty = not self.analysis_state_manager
        previous.add(me)

        found_actions = []

        for key, value in self.__dict__.items():
            if isinstance(value, Action):
                found_actions.append(value)
            elif isinstance(value, (tuple, list, set)):
                for _ in value:
                    if isinstance(_, Action):
                        found_actions.append(_)

        for action in found_actions:
            action.plan(me)
        return previous


class Folder(Action):
    """A Folder action collects a new solution every time the action is called.

    An example of this is the 'post step' for the `xaxis`. A folder action is made
    called `post_step` and is passed to a function which will `do` it multiple times.
    After each step the specificed action is called and its solution will be added to
    the folder.
    """

    def __init__(self, name, action, solution):
        super().__init__(name)
        self.action = action
        self.folder_solution = BaseSolution(name)
        solution.add(self.folder_solution)

    def _do(self, state):
        sol = state.apply(self.action)
        if sol:
            self.folder_solution.add(sol)

    def _requests(self, model, memo, first=True):
        return self.action._requests(model, memo)


class Parallel(Action):
    def __init__(self, *actions):
        super().__init__("parallel", True)
        self.actions = actions

    def _do(self, state):
        sols = []
        for action in self.actions:
            # Need to loop through all the actions that we want to run
            # And build new states to feed into them.
            newstate = state._split()
            if not action.analysis_state_manager:
                # If the next action is managing the state then it should either
                # be building a simulation or passing the state on to something that
                # does. If the next action isn't, like an Xaxis, we should build it
                # so that it can work with it.
                rq = action.get_requests(newstate.model)
                params = tuple(
                    convert_str_to_parameter(newstate.model, _)
                    for _ in rq["changing_parameters"]
                )
                newstate._build_model(params, rq["keep_nodes"])
            sols.append(newstate.apply(action))

        return tuple(sols)

    def _requests(self, model, memo, first=True):
        # Parallel by it's nature has to deepcopy the model
        # that has been passed into it, otherwise there will
        # be all sort of clashes that will be hard to resolve.
        pass


class Series(Action):
    def __init__(self, *actions, flatten=True):
        super().__init__("series", True)
        self.actions = actions
        self.flatten = flatten

    def _do(self, state):
        LOGGER.info(f"Doing action {self}")
        if state.sim is not None:
            # Split the state and work on a new model
            # which will involve rebuilding the current model
            # state and optimising it for whatever this series
            # will be performing
            state = state._split()

        rq = self.get_requests(state.model)
        params = tuple(
            convert_str_to_parameter(state.model, _) for _ in rq["changing_parameters"]
        )
        state._build_model(params, rq["keep_nodes"])

        if self.flatten:
            first = BaseSolution(self.name)
        else:
            first = None

        curr_sol = None
        for i, action in enumerate(self.actions):
            next_sol = state.apply(action)
            if self.flatten and next_sol is not None:
                first.add(next_sol)
            else:
                if next_sol and not curr_sol:
                    first = next_sol  # need to return the first one
                if next_sol:
                    if curr_sol:
                        curr_sol.add(next_sol)
                    curr_sol = next_sol
        return first

    def _requests(self, model, memo, first=True):
        if not first:
            return

        for action in self.actions:
            action._requests(model, memo, False)


class LogModelAttribute(Action):
    def __init__(self, *attrs):
        super().__init__("print_parmeter")
        self.attrs = attrs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        LOGGER.info(*(f"{_}={state.model.reduce_get_attr(str(_))}" for _ in self.attrs))


class Sweep(Action):
    """An action that sweeps N number of parameters through the values in N arrays.

    Parameters
    ----------
    args : [Parameter, str], array, boolean
        Expects 3 arguments per axis. The first is a full name of a Parameter or
        a Parameter object. The second is an array of values to step this
        parameter over, and lastly a boolean value to say whether this is a
        relative step from the parameters initial value.

    pre_step : Action, optional
        An action to perform before the step is computed

    post_step : Action, optional
        An action to perform after the step is computed

    reset_parameter : boolean, optional
        When true this action will reset the all the parameters it changed to
        the values before it ran.

    name : str
        Name of the action, used to find the solution in the final output.
    """

    def __init__(
        self, *args, pre_step=None, post_step=None, reset_parameter=True, name="step"
    ):
        super().__init__(name)
        if len(args) % 3 != 0:
            raise Exception(
                f"Sweep requires triplet of input arguments: parameter, array, relative_change. Not {args}"
            )
        self.args = args
        self.pre_step = pre_step
        self.post_step = post_step
        self.reset_parameter = reset_parameter

        def process_input_parameter(p):
            if isinstance(p, ModelElement):
                if p.default_parameter_name is None:
                    raise ValueError(
                        f"{repr(p)} does not have a default parameter, please specify one to use"
                    )
                p = getattr(p, p.default_parameter_name)

            if isinstance(p, Parameter):
                if not p.changeable_during_simulation:
                    raise Exception(
                        f"Parameter {p.full_name} cannot be changed during a simulation"
                    )
                return p.full_name
            else:
                return p

        self.parameters = tuple(process_input_parameter(p) for p in args[::3])

        self.axes = tuple(np.atleast_1d(_).astype(np.float64) for _ in args[1::3])
        self.offsets = np.array(args[2::3], dtype=np.float64)
        self.out_shape = tuple(np.size(_) for _ in self.axes)

    def _requests(self, model, memo, first=True):
        params = tuple(convert_str_to_parameter(model, _) for _ in self.parameters)
        if self.reset_parameter:
            # Get the actual parameter for this xaxis
            for p in params:
                if p.value is None:
                    raise ValueError(
                        f"Parameters being changed in a simulation must start with a float value not None. Change {repr(p)} to a float value."
                    )

        if any((not p.changeable_during_simulation for p in params)):
            raise Exception(
                f"The property {p.full_name} cannot be changed during a simulation"
            )

        memo["changing_parameters"].extend(self.parameters)
        if self.pre_step:
            self.pre_step._requests(model, memo)
        if self.post_step:
            self.post_step._requests(model, memo)

    def _do(self, state):
        if state.model is None:
            raise Exception("No model was provided")
        if state.sim is None:
            raise Exception("No simulation was provided")

        # Get all the parameters that need to be tuned in this action and
        # any of its pre/post steps
        rq = self.get_requests(state.model)
        all_params = tuple(
            convert_str_to_parameter(state.model, _) for _ in rq["changing_parameters"]
        )
        # Get the actual parameter for this xaxis
        params = tuple(
            convert_str_to_parameter(state.model, _) for _ in self.parameters
        )

        if not all((p.is_tunable for p in all_params)):
            raise Exception(
                f"Not all parameters {params} are tunable in this simulation {state.sim}"
            )

        return self._run(state, *params)

    def _run(self, state, *params):
        # Record intial values of parameters before we go changing
        # anything so we can reset them later
        if self.reset_parameter:
            initial = tuple(
                float(param.value) for param in state.sim.tunable_parameters
            )

        sol = ArraySolution(
            self.name,
            None,
            state.sim.detector_workspaces,
            self.out_shape,
            self.axes,
            params,
        )
        offsets = np.array(params, dtype=np.float64) * self.offsets
        # Make new folder structure in solution if we have any actions
        # that branch off.
        pre_step = Folder("pre_step", self.pre_step, sol) if self.pre_step else None
        post_step = Folder("post_step", self.post_step, sol) if self.post_step else None
        run_axes_scan(
            state, self.axes, params, offsets, self.out_shape, sol, pre_step, post_step,
        )
        if self.reset_parameter:
            # Reset all parameters and if we were changing a geometric parameter
            # reset the beamtrace data to initial state
            for i, param in zip(initial, state.sim.tunable_parameters):
                param.value = i

            # Ensure the __cvalue of each symbolic parameter gets reset accordingly
            for param in state.sim.changing_parameters:
                param._reset_cvalue()

            if any(
                type(p) is GeometricParameter and p.is_symbolic
                for p in state.sim.changing_parameters
            ):
                state.model._update_symbolic_abcds()
            # Need to check all changing parameters incase of symbols
            # if any(type(p) is GeometricParameter for p in state.sim.changing_parameters):
            #    state.model.beam_trace()

        return sol


class Noxaxis(Sweep):
    def __init__(self, pre_step=None, post_step=None, name="noxaxis"):
        super().__init__(name=name, pre_step=None, post_step=None)


class XNaxis(Sweep):
    def __init__(
        self, *args, relative=False, pre_step=None, post_step=None, name="XNaxis"
    ):
        if len(args) % 5 != 0:
            raise Exception(
                f"XNaxis arguments must come in groups of five: parameter, mode, start, stop, steps not {args}"
            )
        self.relative = relative
        self.N = len(args) // 5

        if self.N == 0:
            raise Exception("XNaxis requires at least one axis to be specified")
        # Here we map the XNaxis arguments to the Sweep inputs
        self.__set_args = args
        new_args = []

        for i in range(0, len(args), 5):
            new_args.append(args[i + 0])
            new_args.append(
                get_sweep_array(args[i + 2], args[i + 3], args[i + 4], args[i + 1])
            )
            new_args.append(relative)

        super().__init__(*new_args, pre_step=pre_step, post_step=post_step, name=name)

    def __getattr__(self, key):
        res = re.match("(parameter|mode|start|stop|steps|offset)([0-9]*)", key)
        if res is None:
            super().__getattribute__(key)
        else:
            grp = res.groups()
            N = 1 if grp[1] == "" else int(grp[1])
            if N == 0:
                raise Exception("Specify an axes greater than 0")
            if N > self.N:
                raise Exception(f"This xaxis does not have {N} axes")
            idx = 6 * (N - 1)
            if grp[0] == "parameter":
                return self.__set_args[idx + 0]
            elif grp[0] == "mode":
                return self.__set_args[idx + 1]
            elif grp[0] == "start":
                return self.__set_args[idx + 2]
            elif grp[0] == "stop":
                return self.__set_args[idx + 3]
            elif grp[0] == "steps":
                return self.__set_args[idx + 4]
            elif grp[0] == "offset":
                return self.__set_args[idx + 5]


class Xaxis(XNaxis):
    def __init__(
        self,
        parameter,
        mode,
        start,
        stop,
        steps,
        relative=False,
        pre_step=None,
        post_step=None,
        name="xaxis",
    ):
        super().__init__(
            parameter,
            mode,
            start,
            stop,
            steps,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )


class X2axis(XNaxis):
    def __init__(
        self,
        parameter1,
        mode1,
        start1,
        stop1,
        steps1,
        parameter2,
        mode2,
        start2,
        stop2,
        steps2,
        relative=False,
        pre_step=None,
        post_step=None,
        name="x2axis",
    ):
        super().__init__(
            parameter1,
            mode1,
            start1,
            stop1,
            steps1,
            parameter2,
            mode2,
            start2,
            stop2,
            steps2,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )


class X3axis(XNaxis):
    def __init__(
        self,
        parameter1,
        mode1,
        start1,
        stop1,
        steps1,
        parameter2,
        mode2,
        start2,
        stop2,
        steps2,
        parameter3,
        mode3,
        start3,
        stop3,
        steps3,
        relative=False,
        pre_step=None,
        post_step=None,
        name="x3axis",
    ):
        super().__init__(
            parameter1,
            mode1,
            start1,
            stop1,
            steps1,
            parameter2,
            mode2,
            start2,
            stop2,
            steps2,
            parameter3,
            mode3,
            start3,
            stop3,
            steps3,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )


class RunLocksSolution(BaseSolution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.iters = 0
        self.results = None
        self.lock_names = ()


class RunLocks(Action):
    def __init__(
        self, *locks, exception_on_fail=True, max_iterations=10000, name="run locks"
    ):
        super().__init__(name)
        self.locks = tuple((l if isinstance(l, str) else l.name) for l in locks)
        self.max_iterations = max_iterations
        self.exception_on_fail = exception_on_fail

    def _do(self, state):
        if state.sim is None:
            raise Exception("Simulation has not been built")
        if not isinstance(state.sim, finesse.simulations.CarrierSignalMatrixSimulation):
            raise NotImplementedError()

        recompute = True

        if len(self.locks) == 0:
            locks = state.model.locks
        else:
            locks = tuple(state.model.elements[name] for name in self.locks)

        out_wss = set(  # workspaces can be in both lists
            (*state.sim.readout_workspaces, *state.sim.detector_workspaces)
        )

        dws = tuple(
            next(
                filter(lambda x: x.oinfo.name == lock.error_signal.name, out_wss,),
                None,
            )
            for lock in locks
        )

        N = len(locks)
        # Store initial parameters incase of failure so we can reset the model
        initial_parameters = tuple(float(lock.feedback) for lock in locks)

        sol = RunLocksSolution(self.name)
        sol.iters = -1
        sol.results = np.zeros((len(locks), 2, self.max_iterations + 1))
        sol.lock_names = tuple(lock.name for lock in locks)

        while recompute and sol.iters < self.max_iterations:
            sol.iters += 1
            state.sim.run_carrier()
            recompute = False
            for i in range(N):
                if not locks[i].disabled:
                    acc = locks[i].accuracy
                    res = dws[i].get_output() - locks[i].offset
                    sol.results[i, 0, sol.iters] = res
                    if not (-acc <= res <= acc):
                        # We'll need to recompute the carrier sim
                        recompute = True
                        feedback = locks[i].gain * res
                        locks[i].feedback.value += feedback
                        sol.results[i, 1, sol.iters] = feedback

        if recompute is True:
            if self.exception_on_fail:
                raise Exception("Locks failed: max iterations reached")
            else:
                LOGGER.warn("Locks failed")
                for lock, value in zip(locks, initial_parameters):
                    lock.feedback.value = value
        return sol

    def _requests(self, model, memo, first=True):
        if len(self.locks) == 0:
            # If none given lock everything
            for lock in model.locks:
                memo["changing_parameters"].append(lock.feedback.full_name)
        else:
            for name in self.locks:
                if name not in model.elements:
                    raise Exception(f"Model {model} does not have a lock called {name}")
                memo["changing_parameters"].append(
                    model.elements[name].feedback.full_name
                )


class Scale(Action):
    """Action for scaling simulation outputs by some fixed amount. Included for
    compatibility with legacy Finesse code. New users should apply any desired scalings
    manually from Python.

    Parameters
    ----------
    detectors : dict
        A dictionary of `detector name: scaling factor` mappings.
    """

    def __init__(self, scales: dict, **kwargs):
        super().__init__(None)
        self.kwargs = kwargs
        self.scales = scales

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        sol = state.previous_solution
        for det, fac in self.scales.items():
            sol._outputs[det][()] *= fac


class Debug(Action):
    def __init__(self, name="Debug"):
        super().__init__(name)
        self.cancel = False

    def _requests(self, model, memo, first=True):
        pass

    def do(self, state):
        if not self.cancel:
            from IPython.terminal.embed import InteractiveShellEmbed

            banner = textwrap.dedent(
                f"""
            ---- Finesse Debugging
            Instance          : {self.name}
            Previous solution : s_prev
            Current model     : model
            Current carrier   : carrier
            Current signal    : signal

            To stop future debug calls set : self.cancel = True
            To continue analyis            : exit
            """
            )
            self.shell = InteractiveShellEmbed(banner1=banner)
            self.shell()


### Beam tracing related actions ###


class ABCD(Action):
    """Computation of an ABCD matrix over a path.

    See :func:`.compute_abcd` for details.
    """

    def __init__(self, name="abcd", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.ABCD(solution_name=self.name, **self.kwargs)


class BeamTrace(Action):
    """Full beam tracing on a complete model.

    See :meth:`.Model.beam_trace` for details.
    """

    def __init__(self, name="beam_trace", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.beam_trace(solution_name=self.name, **self.kwargs)


class PropagateBeam(Action):
    """Propagation of a beam, in a single plane, through a given path.

    See :meth:`.Model.propagate_beam` for details.
    """

    def __init__(self, name="propagation", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.propagate_beam(solution_name=self.name, **self.kwargs)


class PropagateAstigmaticBeam(Action):
    """Propagation of a beam, in both planes, through a given path.

    See :meth:`.Model.propagate_beam_astig` for details.
    """

    def __init__(self, name="astig_propagation", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.propagate_beam_astig(solution_name=self.name, **self.kwargs)


class Plot(Action):
    def __init__(self, name="abcd"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        raise NotImplementedError()


class Printer(Action):
    def __init__(self, name="printer"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        raise NotImplementedError()


class PrintModel(Action):
    def __init__(self, name="print_model"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        print(state.model)


class PrintModelAttr(Action):
    def __init__(self, *args):
        super().__init__(self.__class__.__name__)
        self.args = tuple(a.full_name for a in args)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        print(*(f"{_}={state.model.reduce_get_attr(_)}" for _ in self.args))


class FrequencyResponseSolution(BaseSolution):
    def __getitem__(self, key):
        try:
            key = np.atleast_1d(key).tolist()
            inp_key = slice(None, None, None)
            out_key = slice(None, None, None)

            for k in key:
                _k = np.atleast_1d(k)
                if all(_ in self.inputs for _ in _k):
                    inp_key = tuple(self.inputs.index(_) for _ in _k)
                if all(_ in self.outputs for _ in _k):
                    out_key = tuple(self.outputs.index(_) for _ in _k)

            slices = (slice(None, None, None), inp_key, out_key)
            return self.out[slices].squeeze()
        except (ValueError, IndexError, TypeError):
            return super().__getitem__(key)

    def plot_dofs(self, *dofs, axs=None, max_width=12, show_unity=False, **kwargs):
        import matplotlib.pyplot as plt
        import numpy as np

        if len(dofs) == 0:
            dofs = self.inputs

        if axs is None:
            # if no axes are given then grab the figure
            # and any axes that are in it
            fig = plt.gcf()
            axs = np.atleast_2d(fig.axes)
        else:
            axs = np.atleast_2d(axs)
            fig = axs[0, 0].get_figure()

        dofs = np.atleast_1d(dofs)
        N = len(dofs)
        W = min(5, max_width / N)
        if np.prod(axs.shape) != N:
            fig, axs = plt.subplots(
                1, N, figsize=(W * N, 3.5), squeeze=False, sharey=True
            )

        for i, dof in enumerate(dofs):
            axs[0, i].loglog(self.f, abs(self[dof]), **kwargs)
            axs[0, i].legend(self.outputs)
            axs[0, i].set_xlabel("Frequency [Hz]")
            axs[0, i].set_title(dof)
            if show_unity:
                axs[0, i].hlines(
                    1, min(self.f), max(self.f), color="k", ls=":", zorder=-10
                )

        axs[0, 0].set_ylabel("OUTPUT/DOF")
        plt.tight_layout()

        return fig, axs

    plot = plot_dofs  # Default plot option

    def plot_readouts(self, *readouts, axs=None, ls=None, max_width=12):
        import matplotlib.pyplot as plt

        if len(readouts) == 0:
            readouts = self.outputs

        readouts = np.atleast_1d(readouts)
        if axs is None:
            N = len(readouts)
            W = min(5, max_width / N)
            fig, axs = plt.subplots(
                1, N, figsize=(W * N, 3.5), squeeze=False, sharey=True
            )
        else:
            fig = plt.gcf()

        for i, rd in enumerate(readouts):
            axs[0, i].loglog(self.f, abs(self[rd]), ls=ls)
            axs[0, i].legend(self.inputs)
            axs[0, i].set_xlabel("Frequency [Hz]")
            axs[0, i].set_title(rd)

        axs[0, 0].set_ylabel("OUTPUT/DOF")
        plt.tight_layout()

        return fig, axs


class FrequencyResponse(Action):
    """Computes the frequency response of a signal injceted at various nodes to compute
    transfer functions to multiple output nodes. Inputs and outputs should be electrical
    or mechanical nodes. It does this in an efficient way by using the same model and
    solving for multiple RHS input vectors.

    This action does not alter the model state.

    Parameters
    ----------
    f : array, double
        Frequencies to compute the transfer functions over
    inputs : iterable[str or Element]
        Mechanical or electrical node to inject signal at
    outputs : iterable[str or Element]
        Mechanical or electrical nodes to measure output at
    open_loop : bool, optional
        Computes open loop transfer functions if the system has closed
    name : str, optional
        Solution name

    Examples
    --------
    Here we measure a set of transfer functions from DARM and CARM
    to four readouts for a particular `model`,

    >>> sol = FrequencyResponse(np.geomspace(0.1, 50000, 100),
    ...         ('DARM', 'CARM'),
    ...         ('AS.DC', 'AS45.I', 'AS45.Q', 'REFL9.I'),
    ... ).run(model)

    Single inputs and outputs can also be specified

    >>> FrequencyResponse(np.geomspace(0.1, 50000, 100), 'DARM', AS.DC').run(model)

    The transfer functions can then be accessed like a 2D array by name,
    the ordering of inputs to outputs does not matter.

    >>> sol['DARM'] # DARM to all outputs
    >>> sol['DARM', 'AS.DC'] # DARM to AS.DC
    >>> sol['DARM', ('AS.DC', 'AS45.I')]
    >>> sol['AS.DC'] # All inputs to AS.DC readout
    """

    def __init__(self, f, inputs, outputs, *, open_loop=False, name="inject"):
        super().__init__(name)
        inputs = np.atleast_1d(inputs)
        outputs = np.atleast_1d(outputs)

        try:
            self.f = np.array(f, dtype=np.float64, copy=True)
        except Exception:
            # If the f is a symbol...
            self.f = np.array(f.eval(), dtype=np.float64, copy=True)

        def process(x):
            if not isinstance(x, (str, np.str_)):
                return x.full_name
            return x

        self.inputs = list(process(i) for i in inputs)
        self.outputs = list(process(o) for o in outputs)
        self.open_loop = open_loop

    def _do(self, state, fsig_independant_outputs=None, fsig_dependant_outputs=None):
        input_rhs_indices = np.zeros(len(self.inputs), dtype=int)
        output_rhs_indices = np.zeros(len(self.outputs), dtype=int)

        for i, node in enumerate(state.name_to_elec_mech_nodes(self.inputs)):
            input_rhs_indices[i] = state.sim.signal.field(node, 0, 0)
        for i, node in enumerate(state.name_to_elec_mech_nodes(self.outputs)):
            output_rhs_indices[i] = state.sim.signal.field(node, 0, 0)

        sol = FrequencyResponseSolution(self.name)
        sol.f = self.f
        sol.inputs = self.inputs
        sol.outputs = self.outputs
        state.sim.carrier.run()
        rtn = run_fsig_sweep(
            state.sim,
            self.f,
            input_rhs_indices,
            output_rhs_indices,
            None,
            self.open_loop,
            tuple(fsig_independant_outputs)
            if fsig_independant_outputs is not None
            else None,
            tuple(fsig_dependant_outputs)
            if fsig_dependant_outputs is not None
            else None,
        )
        if len(rtn) == 2:
            sol.out = rtn[0]
            sol.extra_outputs = rtn[1]
        else:
            sol.out = rtn

        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append("fsig.f")
        memo["keep_nodes"].extend(self.inputs)
        memo["keep_nodes"].extend(self.outputs)


class OptimiseRFReadoutPhaseDCSolution(BaseSolution):
    pass


class OptimiseRFReadoutPhaseDC(Action):
    """This optimises the demodulation phase of ReadoutRF elements relative to some
    DegreeOfFreedom. This optimises the phases so that the ReadoutRF in-phase signal
    will optimally sense the provided DegreeOfFreedom.

    The phases are optimised by calculating the DC response of the readouts.

    This Action changes the state of the model.

    Parameters
    ----------
    args
        Pairs of DegreesOfFreedom and ReadoutRF elements, or pairs of their names.
    d_dof : float, optional
        The small offset applied to the DOFs to compute the gradients of the error
        signals.

    Examples
    --------
    Here we optimise REFL9 I and AS45 I to sense CARM and DARM optimially:
    >>> sol = OptimiseRFReadoutPhaseDC("CARM", "REFL9", "DARM", "AS45").run(aligo)
    """

    def __init__(self, *args, d_dof=1e-9, name="optimise_demod_phases_dc"):
        super().__init__(name)
        self.args = args
        self.dofs = args[::2]
        self.readouts = args[1::2]
        self.d_dof = d_dof

        if len(self.dofs) != len(self.readouts):
            raise ValueError(
                "Pairs of Degrees of freedoms and readouts must be provided"
            )

    def _do(self, state):
        Idws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_I", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        Qdws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_Q", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        dcs = tuple(state.model.reduce_get_attr(f"{dof}.DC") for dof in self.dofs)

        N = len(self.dofs)
        sol = OptimiseRFReadoutPhaseDCSolution(self.name)
        sol.Ivals = np.zeros((N, 2), dtype=complex)
        sol.Qvals = np.zeros((N, 2), dtype=complex)
        # Here we compute the gradient of the error signals
        # with respect to some DOF change
        for i in range(N):
            dcs[i].value -= self.d_dof
            state.sim.run_carrier()
            sol.Ivals[i, 0] = Idws[i].get_output()
            sol.Qvals[i, 0] = Qdws[i].get_output()
            dcs[i].value += 2 * self.d_dof
            state.sim.run_carrier()
            sol.Ivals[i, 1] = Idws[i].get_output()
            sol.Qvals[i, 1] = Qdws[i].get_output()
            # reset value
            dcs[i].value -= self.d_dof
        # Compute the gradients in both I and Q
        sol.Igradients = (sol.Ivals[:, 1] - sol.Ivals[:, 0]) / 2e-6
        sol.Qgradients = (sol.Qvals[:, 1] - sol.Qvals[:, 0]) / 2e-6
        # We can use the complex angle to compute how much to change the
        # demod phase by to optimise it
        sol.add_degrees = np.angle(sol.Igradients + 1j * sol.Qgradients, deg=True)
        sol.phases = {}
        for i in range(N):
            param = state.model.reduce_get_attr(f"{self.readouts[i]}.phase")
            param.value += sol.add_degrees[i]
            sol.phases[self.readouts[i]] = float(param.value)

        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].extend((f"{_}.DC" for _ in self.dofs))
        memo["changing_parameters"].extend((f"{_}.phase" for _ in self.readouts))
        return memo


class SensingMatrixDCSolution(BaseSolution):
    """Sensing matrix DC solution.

    The raw sensing matrix information can be accessed using the
    `SensingMatrixDCSolution.out` member. This is a complex-valued array with dimensions
    (DOFs, Readouts), which are accessible via `SensingMatrixDCSolution.dofs` and
    `SensingMatrixDCSolution.readouts`.

    A table can be printed using :meth:`.SensingMatrixDCSolution.display`.

    Polar plot can be generated using :meth:`.SensingMatrixDCSolution.plot`

    Printing :class:`.SensingMatrixDCSolution` will show an ASCII table of the data.
    """

    def display(self, dofs=None, readouts=None, tablefmt="html", floatfmt=".2G"):
        """Displays a HTML table of the sensing matrix.

        Notes
        -----
        Only works when called from an IPython environmeny with the `display`
        method available.

        Parameters
        ----------
        dofs : iterable[str], optional
            Names of degrees of freedom to show, defaults to all if None
        readouts : iterable[str], optional
            Names of readouts to show, defaults to all if None
        """
        from IPython import display

        B, dofs, readouts = self.matrix_data(dofs, readouts)
        if tablefmt == "html":
            display(
                tabulate(
                    B,
                    headers=readouts,
                    showindex=dofs,
                    tablefmt=tablefmt,
                    floatfmt=floatfmt,
                )
            )
        else:
            print(
                tabulate(
                    B,
                    headers=readouts,
                    showindex=dofs,
                    tablefmt=tablefmt,
                    floatfmt=floatfmt,
                )
            )

    def __str__(self):
        B, dofs, readouts = self.matrix_data()
        return tabulate(
            B, headers=readouts, showindex=dofs, tablefmt="fancy_grid", floatfmt=".2G"
        )

    def matrix_data(self, dofs=None, readouts=None):
        """Generates a sensing matrix table.

        Parameters
        ----------
        dofs : iterable[str], optional
            Names of degrees of freedom to show, defaults to all if None
        readouts : iterable[str], optional
            Names of readouts to show, defaults to all if None

        Returns
        -------
        matrix : 2D numpy array, complex
        dofs : list of :class:`str`
        readouts: list of :class:`str`
        """
        dofs = dofs or self.dofs
        readouts = readouts or self.readouts
        dofs = np.atleast_1d(dofs)
        readouts = np.atleast_1d(readouts)
        hdrs = tuple(rd + iq for rd in readouts for iq in ("_I", "_Q"))
        sl1 = tuple(dofs.index(_) for _ in dofs)
        sl2 = tuple(readouts.index(_) for _ in readouts)
        # Reshaping so that we have extra columns with I and Q signals
        A = self.out[sl1, :][:, sl2]
        Nr, Nc = A.shape
        B = np.zeros(2 * Nr * Nc)
        B[0::2] = A.real.flat
        B[1::2] = A.imag.flat
        B = B.reshape(Nr, 2 * Nc)
        return B, dofs.tolist(), hdrs

    def plot(self, Nrows, Ncols, figsize=(6, 5), *, dofs=None, readouts=None):
        import matplotlib.pyplot as plt

        dofs = np.atleast_1d(dofs or self.dofs)
        readouts = np.atleast_1d(readouts or self.readouts)

        fig, axs = plt.subplots(
            Nrows,
            Ncols,
            figsize=figsize,
            subplot_kw={"projection": "polar"},
            squeeze=False,
        )
        axs = axs.flatten()
        for idx in range(len(readouts)):
            dof_idxs = tuple(self.dofs.index(_) for _ in dofs)
            _ax = axs[idx]
            A = self.out[dof_idxs, idx]

            _ax.set_theta_zero_location("E")
            r_lim = (np.log10(np.abs(A)).min() - 1, np.log10(np.abs(A)).max())
            _ax.set_ylim(r_lim[0], r_lim[1] + 1)
            _ax.set_yticklabels([])

            theta = np.angle(A)
            r = np.log10(np.abs(A))
            _ax.plot(
                (theta, theta),
                (r_lim[0] * np.ones_like(r), r),
                marker="D",
                markersize=5,
            )
            _ax.set_title(self.readouts[idx])
        _ax.legend(self.dofs, loc="best", bbox_to_anchor=(0.5, -0.3), fontsize=8)
        plt.tight_layout(pad=1.2)
        return fig, axs


class SensingMatrixDC(Action):
    """Computes the sensing matrix elements for various degrees of freedom and readouts
    that should be present in the model. The solution object for this action then
    contains all the information on the sensing matrix. This can be plotted in polar
    coordinates, displayed in a table, or directly accessed.

    The sensing gain is computed by calculating the gradient of each readout
    signal, which means it is a DC measurement. This will not include any
    suspension or radiation pressure effects.

    This action does not modify the states model.

    Parameters
    ----------
    dofs : iterable[str]
        String names of degrees of freedom
    readouts : iterable[str]
        String names of readouts
    d_dof : float, optional
        Small step used to compute derivative
    """

    def __init__(self, dofs, readouts, d_dof=1e-9, name="sensing_matrix_dc"):
        super().__init__(name)
        self.dofs = dofs
        self.readouts = readouts
        self.d_dof = d_dof

    def _do(self, state):
        Idws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_I", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        Qdws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_Q", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        dcs = tuple(state.model.reduce_get_attr(f"{dof}.DC") for dof in self.dofs)

        Nd = len(self.dofs)
        Nr = len(self.readouts)

        sol = SensingMatrixDCSolution(self.name)
        sol.dofs = self.dofs
        sol.readouts = self.readouts
        sol.Ivals = np.zeros((Nd, Nr, 2), dtype=float)
        sol.Qvals = np.zeros((Nd, Nr, 2), dtype=float)
        # Here we compute the gradient of the error signals
        # with respect to some DOF change
        for i in range(Nd):
            dcs[i].value -= self.d_dof
            state.sim.run_carrier()
            for j in range(Nr):
                sol.Ivals[i, j, 0] = Idws[j].get_output()
                sol.Qvals[i, j, 0] = Qdws[j].get_output()
            dcs[i].value += 2 * self.d_dof
            state.sim.run_carrier()
            for j in range(Nr):
                sol.Ivals[i, j, 1] = Idws[j].get_output()
                sol.Qvals[i, j, 1] = Qdws[j].get_output()
            # reset value
            dcs[i].value -= self.d_dof
        # Compute the gradients in both I and Q
        sol.Igradients = (sol.Ivals[:, :, 1] - sol.Ivals[:, :, 0]) / 2e-6
        sol.Qgradients = (sol.Qvals[:, :, 1] - sol.Qvals[:, :, 0]) / 2e-6
        sol.out = sol.Igradients + 1j * sol.Qgradients
        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].extend((f"{_}.DC" for _ in self.dofs))
        return memo


class Change(Action):
    """Changes a model Parameter to some value during an analysis."""

    def __init__(self, change_dict=None, *, relative=False, **kwargs):
        super().__init__(None)
        self.kwargs = kwargs or {}
        if change_dict:
            self.kwargs.update(change_dict)
        self.relative = relative

    def _requests(self, model, memo, first=True):
        from finesse import Parameter

        for el in self.kwargs.keys():
            p = convert_str_to_parameter(model, el)
            if isinstance(p, Parameter):
                memo["changing_parameters"].append(el)
            else:
                raise TypeError(
                    f"{el} is not a name of a Parameter or Component in the model"
                )

    def _do(self, state):
        for el, val in self.kwargs.items():
            p = convert_str_to_parameter(state.model, el)
            if self.relative:
                p.value += val
            else:
                p.value = val


class NoiseProjectionSolution(BaseSolution):
    def plot(self, output_node=None, lower=0.1, upper=3, *, ax=None, **kwargs):
        import matplotlib.pyplot as plt

        if output_node is None:
            output_node = self.output_nodes[0]

        if ax is None:
            fig = plt.gcf()
            if len(fig.axes) == 0:
                fig.subplots(1, 1)
            ax = fig.axes[0]

        total = np.sqrt((self.out[output_node] ** 2).sum(1))
        rng = lower * total.min(), upper * total.max()
        noises_to_plot = np.any(self.out[output_node] > rng[0], 0)
        ax.loglog(self.f, self.out[output_node][:, noises_to_plot])
        ax.loglog(self.f, total, c="k", ls="-.", lw=2)
        ax.legend((*np.array(self.noises)[noises_to_plot], "Total"))
        ax.set_ylim(*rng)
        ax.set_ylabel(
            f"ASD [{output_node if not self.scaling else self.scaling}/$\\sqrt{{\\mathrm{{Hz}}}}$]"
        )
        ax.set_xlabel("Frequency [Hz]")


class NoiseProjection(Action):
    def __init__(self, f, *output_nodes, scaling=None, name="loop"):
        if len(output_nodes) == 0:
            raise ValueError(
                "At least one output node must be specified to compute noise projection to"
            )
        super().__init__(name)
        process = lambda x: x.full_name if type(x) is not str else x

        self.f = f
        self.scaling = process(scaling) if scaling is not None else None
        self.output_nodes = tuple(process(o) for o in output_nodes)

        if len(self.output_nodes) > len(set(self.output_nodes)):
            raise ValueError(
                f"The same output node has been requested multiple times {self.output_nodes}"
            )

    def _do(self, state):
        sol = NoiseProjectionSolution(self.name)
        sol.f = self.f
        sol.output_nodes = self.output_nodes
        sol.scaling = self.scaling
        # create a list of callables func(fsig) to get the ASD noises
        noise_ASDs = {
            name: el.ASD.lambdify(state.model.fsig.f)
            for name, el in state.model.noises.items()
        }
        # labels for noises
        sol.noises = list(el.name for _, el in state.model.noises.items())
        # Keep track of which nodes have what noise injected into them
        noise_node_map = defaultdict(list)
        for name, el in state.model.noises.items():
            noise_node_map[el.node.full_name].append(el.name)
        # Collect any extra outputs that should be calculated during the fsig sweep. This
        # is to make efficient use of the filling and solving that this is already doing
        # to extract quantum noise, or others, as needed. Some of these outputs will be
        # signal frequency independant, such as standard shot-noise calculations, so just
        # compute them once.
        # TODO eventually handle qnoised detectors, which are frequency dependant
        fsig_indep_output = []
        for dws in state.sim.readout_workspaces:
            if isinstance(dws, (QShot0Workspace, QShotNWorkspace)):
                added = 0  # don't calculate anything if we aren't modelling the nodes
                # The quantum shot noise detectors will be on the optiical
                # input node, which we can't inject a signal into for computing
                # the noise propagation. Here we need to get the electrical outputs
                # of the readout and just put the noise in there
                for n in dws.owner.electrical_nodes:
                    if n.full_name in state.sim.signal.nodes:
                        noise_node_map[n.full_name].append(dws.oinfo.name)
                        added += 1
                if added:
                    fsig_indep_output.append(dws)
        # Compute all the required transfer functions for noise propagation
        # NOTE: We use _do directly here because we just want to call the action
        # on this state, rather than `run` which will try and create a new state.
        # This is fine, as long as we have requested all the options it needs in
        # _requests. We can't make this frequency response in the init as we do
        # not have the model to grab all the various noise and shot-noise nodes
        self.input_nodes = tuple(noise_node_map.keys())
        sol.freqresp = FrequencyResponse(
            self.f, self.input_nodes, self.output_nodes
        )._do(state, fsig_indep_output)
        # Get any shot noise outputs from the solution
        for dws in fsig_indep_output:
            # Make a simple callable to work wiht the noise ASD functions
            noise_ASDs[dws.oinfo.name] = lambda f: sol.freqresp.extra_outputs[
                dws.oinfo.name
            ]
            sol.noises.append(dws.oinfo.name)
        # get a map from nodes->noise, for noise->node index in output
        inv_noise_node_map = {}
        for k, v in noise_node_map.items():
            for n in v:
                inv_noise_node_map[n] = self.input_nodes.index(k)
        # Convert all the ASDs into PSDs
        sol.PSDs = np.array(
            tuple(np.ones_like(self.f) * fn(self.f) ** 2 for fn in noise_ASDs.values())
        ).T
        # Use this to broadcast from the frequency response output to get the right
        # transfer function for each noise source
        out_indices = tuple(inv_noise_node_map[name] for name in noise_ASDs.keys())
        # Here we can compute some projection for calculating equivalent noise budgets
        if self.scaling:
            sol.scaling_solution = FrequencyResponse(
                self.f, self.scaling, self.output_nodes, open_loop=True
            )._do(state)
        sol.out = {}
        # compute abs(H)**2 for noise projection of PSDS
        HH = np.zeros(
            (len(self.f), len(out_indices), len(self.output_nodes)), dtype=float
        )
        np.abs(sol.freqresp.out[:, out_indices, :], out=HH)
        np.multiply(HH, HH, out=HH)
        # The final index of HH is the output node index, so we can quickly iterate over
        # them here to project the noises
        for i, output_node in enumerate(self.output_nodes):
            # sqrt(output**2/node**2 * node**2/Hz) => output/rtHz
            sol.out[output_node] = np.sqrt(HH[:, :, i] * sol.PSDs)
            if self.scaling:
                # output / scaling
                # sqrt(output**2/node**2 * node**2/Hz) => scaling/rtHz
                sol.out[output_node] /= np.abs(sol.scaling_solution.out[:, :, 0])
        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append("fsig.f")
        memo["keep_nodes"].extend(self.output_nodes)
        if self.scaling:
            memo["keep_nodes"].append(self.scaling)
        memo["keep_nodes"].extend((el.node.full_name for n, el in model.noises.items()))


##########################################################################################
##########################################################################################
##########################################################################################
##########################################################################################

# class BeamTrace(Action):
#     """Action for tracing the beam throughout an entire model."""

#     def __init__(self, name, **kwargs):
#         super().__init__(name)
#         self.kwargs = kwargs
#         self._info.makes_solution = True

#     def do(self, ws):
#         ws.s_prev.add(ws.model.beam_trace(solution_name=self.name, **self.kwargs))


# class ABCD(Action):
#     """Action to compute a composite ABCD matrix over a given path of a model."""

#     def __init__(self, name, **kwargs):
#         super().__init__(name)
#         self.kwargs = kwargs
#         self._info.makes_solution = True

#     def do(self, ws):
#         ws.s_prev.add(ws.model.ABCD(solution_name=self.name, **self.kwargs))


# class Plot(Action):
#     def __init__(self, *args, **kwargs):
#         super().__init__(self.__class__.__name__)
#         self.args = args
#         self.kwargs = kwargs

#     def do(self, s_prev, model):
#         while type(s_prev) is BaseSolution:
#             s_prev = s_prev.parent

#         if s_prev is not None and hasattr(s_prev, "plot"):
#             s_prev.plot()
#         else:
#             print(f"No plot method found in {s_prev}")


# class Printer(Action):
#     def __init__(self):
#         super().__init__(self.__class__.__name__,)

#     def do(self, ws):
#         s_prev, model = ws.s_prev, ws.model
#         print(s_prev, model)


# class PrintModel(Action):
#     def __init__(self):
#         super().__init__(self.__class__.__name__)

#     def do(self, ws):
#         print(ws.model)


# class PrintSolution(Action):
#     def __init__(self):
#         super().__init__(self.__class__.__name__)

#     def do(self, ws):
#         print(ws.s_prev)


# class PrintAttr(Action):
#     def __init__(self, *args):
#         super().__init__(self.__class__.__name__)
#         self.args = args

#     def do(self, ws):
#         print(*(f"{_}={ws.model.reduce_get_attr(_)}" for _ in self.args))


# class ReprAttr(Action):
#     def __init__(self, *args):
#         super().__init__(self.__class__.__name__)
#         self.args = args

#     def do(self, ws):
#         print(*(f"{_}={repr(ws.model.reduce_get_attr(_))}" for _ in self.args))



"""Kat script generator."""

import logging
from io import StringIO
from functools import singledispatchmethod
from collections.abc import Iterable

import numpy as np

from ..model import Model
from ..element import ModelElement
from ..parameter import Parameter, ParameterRef
from ..symbols import OPERATORS, FUNCTIONS, Operation, Constant
from ..components import Port, Node
from ..analysis.actions import Action
from .containers import (
    KatToken,
    KatNumberToken,
    KatStringToken,
    KatWhitespaceToken,
    KatNoneToken,
    KatScript,
    KatElement,
    KatFunction,
    KatKwarg,
    KatExpression,
    KatGroupedExpression,
    KatArray,
)
from .tokens import LITERALS
from .spec import KatSpec
from .graph import KatNodeType, KatEdgeType, KatGraph
from .util import scriptsorted, merge_attributes


LOGGER = logging.getLogger(__name__)


class ElementContainer:
    """Container for top level model elements, used by the generator."""

    def __init__(self, element):
        super().__init__()
        self.element = element

    def __repr__(self):
        return f"ElementContainer({repr(self.element)})"


class CommandContainer:
    """Container for top level commands, used by the generator."""

    def __init__(self, adapter, *args, **kwargs):
        super().__init__()
        self.adapter = adapter
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return f"CommandContainer({self.adapter.full_name})"


class KatUnbuilder:
    """Model to kat script."""

    # Model element types not dumped.
    IGNORED_ELEMENTS = [
        "Fsig",  # Models always contain an Fsig. The actual value is dumped as a command.
    ]

    # FIXME: move to KatSpec.
    UNARY_OPERATORS = {
        "pos": "+",
        "neg": "-",
    }

    # Reverse dicts for various types.
    LITERAL_MAP = dict((value, key) for key, value in LITERALS.items())

    def __init__(self, spec=None):
        if spec is None:
            spec = KatSpec()

        self.spec = spec
        self.graph = None
        self._lineno = None
        self._index = None

    def unbuild(self, item):
        fobj = StringIO()
        self.unbuild_file(fobj, item)
        fobj.seek(0)
        return fobj.read()

    def unbuild_file(self, fobj, item):
        self.graph = KatGraph()
        self._lineno = 1
        self._index = 1
        self._fill(item, self.graph.ROOT_NODE_NAME)

        unfiller = KatUnfiller()
        unfiller.unfill_file(fobj, self.graph.ROOT_NODE_NAME, self.graph)

    def _fill(self, value, path, **attributes):
        LOGGER.debug(f"filling {value!r} ({type(value)}) into graph at {path}")

        assert path not in self.graph

        attributes = merge_attributes(
            attributes, self._item_node_attributes(value, path)
        )
        self.graph.add_node(path, **attributes)

        return attributes

    @singledispatchmethod
    def _item_node_attributes(self, item, path):
        """Get the attributes for `item` to be added to the corresponding node."""
        raise NotImplementedError(
            f"don't know how to handle an item with type '{item.__class__.__name__}'"
        )

    @_item_node_attributes.register(Model)
    def _(self, model, path):
        extra_tokens = []
        order = 0

        # Components, detectors, etc.
        for element in model.elements.values():
            element_type = element.__class__.__name__
            if element_type in self.IGNORED_ELEMENTS:
                LOGGER.debug(f"skipping filling of ignored element type {element_type}")
                continue

            if order > 0:
                extra_tokens.append(self._newline_token())

            # Define a proxy so we can differentiate model element definitions from
            # references to model elements.
            proxy = ElementContainer(element)
            element_path = self.graph.item_node_name(order, path)
            if self._fill(proxy, element_path):
                self.graph.add_edge(
                    element_path, path, type=KatEdgeType.ARGUMENT, order=order
                )

            order += 1

        # Commands.
        for adapter in self.spec.commands.values():
            command_params = adapter.get(model)

            if command_params is None:
                # Nothing to dump.
                continue

            if adapter.singular:
                # The getter should return a single args, kwargs tuple.
                command_params = [command_params]

            # There are command(s) to dump for this adapter.
            for args, kwargs in command_params:
                if order > 0:
                    extra_tokens.append(self._newline_token())

                proxy = CommandContainer(adapter, *args, **kwargs)
                command_path = self.graph.item_node_name(order, path)
                if self._fill(proxy, command_path):
                    self.graph.add_edge(
                        command_path, path, type=KatEdgeType.ARGUMENT, order=order
                    )

                order += 1

        # Analysis.
        if model.analysis:
            if order > 0:
                extra_tokens.append(self._newline_token())

            analysis_path = self.graph.item_node_name(order, path)
            if self._fill(model.analysis, analysis_path):
                self.graph.add_edge(
                    analysis_path, path, type=KatEdgeType.ARGUMENT, order=order
                )

            order += 1

        return {"type": KatNodeType.ROOT, "extra_tokens": extra_tokens}

    @_item_node_attributes.register(ElementContainer)
    def _(self, proxy, path):
        """Model element definitions."""
        element = proxy.element
        adapter = self.spec.adapter_by_setter(type(element))
        params = adapter.get(element)
        if not params:
            return

        name, args, kwargs = params

        extra_tokens = []
        token = self._name_token(adapter.full_name)
        extra_tokens.append(self._space_token())
        name_token = self._name_token(name)
        extra_tokens.append(self._space_token())
        extra_tokens.extend(self._fill_element_args(args, kwargs, path))

        return {
            "token": token,
            "name_token": name_token,
            "type": KatNodeType.ELEMENT,
            "extra_tokens": extra_tokens,
        }

    @_item_node_attributes.register(CommandContainer)
    def _(self, proxy, path):
        """Command definitions."""
        return self._fill_function(proxy.adapter, proxy.args, proxy.kwargs, path)

    @_item_node_attributes.register(Action)
    def _(self, action, path):
        adapter = self.spec.adapter_by_setter(type(action))
        params = adapter.get(action)
        if not params:
            return
        return self._fill_function(adapter, *params, path)

    @_item_node_attributes.register(Parameter)
    def _(self, parameter, path):
        return self._fill(parameter.value, path)

    @_item_node_attributes.register(ParameterRef)
    def _(self, reference, path):
        value = self._name_token(f"&{reference.name}")
        return {"token": value, "type": KatNodeType.PARAMETER_REFERENCE}

    @_item_node_attributes.register(ModelElement)
    def _(self, value, path):
        """Reference to a model element.

        Model element definitions are matched as :class:`.ElementContainer`.
        """
        return {"token": self._name_token(value.name), "type": KatNodeType.VALUE}

    @_item_node_attributes.register(Port)
    @_item_node_attributes.register(Node)
    def _(self, value, path):
        return {"token": self._name_token(value.full_name), "type": KatNodeType.VALUE}

    @_item_node_attributes.register(int)
    @_item_node_attributes.register(np.integer)
    def _(self, value, path):
        value = int(value)
        if value < 0:
            return self._fill(FUNCTIONS["neg"](abs(value)), path)
        return {"token": self._number_token(value), "type": KatNodeType.VALUE}

    @_item_node_attributes.register(float)
    @_item_node_attributes.register(np.floating)
    def _(self, value, path):
        value = float(value)
        if value < 0:
            return self._fill(FUNCTIONS["neg"](abs(value)), path)
        return {"token": self._number_token(value), "type": KatNodeType.VALUE}

    @_item_node_attributes.register(complex)
    def _(self, value, path):
        real = value.real
        imag = value.imag

        if not real:
            cplx = complex(f"{abs(imag)}j")
            if imag < 0:
                return self._fill(FUNCTIONS["neg"](cplx), path)
            return {"token": self._number_token(cplx), "type": KatNodeType.VALUE}
        elif not imag:
            if real < 0:
                return self._fill(FUNCTIONS["neg"](abs(real)), path)
            return {"token": self._number_token(real), "type": KatNodeType.VALUE}
        else:
            if real < 0:
                real = FUNCTIONS["neg"](abs(real))
            binop = OPERATORS["__add__"] if imag >= 0 else OPERATORS["__sub__"]
            return self._fill(binop(real, self._number_token(f"{abs(imag)}j")), path)

    @_item_node_attributes.register(str)
    def _(self, value, path):
        if value in self.spec.keywords:
            value = self._name_token(value)
        else:
            value = self._string_token(value)
        return {"token": value, "type": KatNodeType.VALUE}

    @_item_node_attributes.register(bool)
    @_item_node_attributes.register(np.bool_)
    def _(self, value, path):
        return {
            "token": self._name_token("true" if value else "false"),
            "type": KatNodeType.VALUE,
        }

    @_item_node_attributes.register(type(None))
    def _(self, value, path):
        return {"token": self._none_token(), "type": KatNodeType.VALUE}

    @_item_node_attributes.register(Iterable)
    def _(self, array, path):
        extra_tokens = []
        extra_tokens.append(self._literal_token("["))
        extra_tokens.extend(self._fill_function_args(array, {}, path))
        extra_tokens.append(self._literal_token("]"))
        return {"type": KatNodeType.ARRAY, "extra_tokens": extra_tokens}

    @_item_node_attributes.register(Constant)
    def _(self, constant, path):
        return self._fill(constant.eval(), path)

    @_item_node_attributes.register(Operation)
    def _(self, operation, path):
        op = operation.name
        extra_tokens = []

        def make_arg(argument, order):
            argument_path = self.graph.item_node_name(order, path)
            self._fill(argument, argument_path)
            self.graph.add_edge(
                argument_path, path, type=KatEdgeType.ARGUMENT, order=order
            )

        if op in self.spec.binary_operators:
            assert len(operation.args) == 2
            extra_tokens.append(self._literal_token("("))
            make_arg(operation.args[0], 0)
            token = self._literal_token(operation.name)
            nodetype = KatNodeType.EXPRESSION
            make_arg(operation.args[1], 1)
            extra_tokens.append(self._literal_token(")"))
        elif unary_op := self.UNARY_OPERATORS.get(op):
            token = self._literal_token(unary_op)
            nodetype = KatNodeType.FUNCTION
            make_arg(operation.args[0], 0)
        elif op in self.spec.expression_functions:
            token = self._name_token(operation.name)
            nodetype = KatNodeType.FUNCTION
            extra_tokens.append(self._literal_token("("))
            for order, argument in enumerate(operation.args):
                if order > 0:
                    extra_tokens.append(self._literal_token(","))
                    extra_tokens.append(self._space_token())
                make_arg(argument, order)
            extra_tokens.append(self._literal_token(")"))
        else:
            raise NotImplementedError(f"don't know how to handle operation '{op}'")

        return {"token": token, "type": nodetype, "extra_tokens": extra_tokens}

    def _token(self, type, raw_value, tokcls=None):
        if tokcls is None:
            tokcls = KatToken

        start_index = self._index
        self._index += len(raw_value)
        stop_index = self._index

        return tokcls(self._lineno, start_index, stop_index, type, raw_value)

    def _name_token(self, value):
        return self._token("NAME", value)

    def _space_token(self, length=1):
        return self._token("WHITESPACE", " " * length, tokcls=KatWhitespaceToken)

    def _number_token(self, value):
        return self._token("NUMBER", str(value), tokcls=KatNumberToken)

    def _string_token(self, value):
        return self._token("STRING", repr(value), tokcls=KatStringToken)

    def _none_token(self):
        return self._token("NONE", "none", tokcls=KatNoneToken)

    def _literal_token(self, literal):
        return self._token(self.LITERAL_MAP[literal], literal)

    def _newline_token(self, count=1):
        self._lineno += count
        self._index = 1
        return self._token("NEWLINE", "\n" * count)

    def _fill_arg(self, value, order, path, **attributes):
        argument_path = self.graph.item_node_name(order, path)
        self._fill(value, argument_path, **attributes)
        self.graph.add_edge(argument_path, path, type=KatEdgeType.ARGUMENT, order=order)

    def _fill_args(self, args, kwargs, path, delimfunc):
        extra_tokens = []

        # Can't use enumerate because we loop twice.
        order = 0

        for value in args:
            if order > 0:
                extra_tokens.extend(delimfunc())
            self._fill_arg(value, order, path)
            order += 1

        for key, value in kwargs.items():
            if order > 0:
                extra_tokens.extend(delimfunc())
            key_token = self._name_token(key)
            equals_token = self._literal_token("=")
            self._fill_arg(
                value, order, path, key_token=key_token, extra_tokens=[equals_token]
            )
            order += 1

        return extra_tokens

    def _fill_element_args(self, args, kwargs, path):
        def delim():
            return [self._space_token()]

        return self._fill_args(args, kwargs, path, delim)

    def _fill_function_args(self, args, kwargs, path):
        def delim():
            return [self._literal_token(","), self._space_token()]

        return self._fill_args(args, kwargs, path, delim)

    def _fill_function(self, adapter, args, kwargs, path):
        extra_tokens = []
        token = self._name_token(adapter.full_name)
        extra_tokens.append(self._literal_token("("))
        extra_tokens.extend(self._fill_function_args(args, kwargs, path))
        extra_tokens.append(self._literal_token(")"))

        return {
            "token": token,
            "type": KatNodeType.FUNCTION,
            "extra_tokens": extra_tokens,
        }


class KatUnfiller:
    """KatGraph to kat script."""

    def unfill(self, node, graph):
        fobj = StringIO()
        self.unfill_file(fobj, node, graph)
        fobj.seek(0)
        return fobj.read()

    def unfill_file(self, fobj, node, graph):
        production = self.production(node, graph)
        unparser = KatUnparser()
        unparser.unparse_file(fobj, production)

    def production(self, node, graph):
        data = graph.nodes[node]
        nodetype = data["type"]

        # Create any dependent arguments.
        arguments = scriptsorted(
            [
                self.production(argument_node, graph)
                for argument_node in graph.dependent_argument_nodes(node)
            ]
        )
        # Grab any extra tokens.
        extra = scriptsorted(data.get("extra_tokens", []))

        # Detect kwargs.
        kwarg_extra = None
        if "key_token" in data:
            # Extract the equals token out of extra.
            equals = next(filter(lambda item: item.type == "EQUALS", extra))
            extra.remove(equals)
            kwarg_extra = [equals]

        if nodetype in KatNodeType.GENERATOR_TERMINAL_NODES:
            value = data["token"]
        elif nodetype == KatNodeType.ROOT:
            value = KatScript(arguments=arguments, extra=extra)
        elif nodetype == KatNodeType.ELEMENT:
            value = KatElement(
                directive=data["token"],
                arguments=arguments,
                extra=extra,
                name=data.get("name_token"),
            )
        elif nodetype == KatNodeType.FUNCTION:
            value = KatFunction(
                directive=data["token"], arguments=arguments, extra=extra
            )
        elif nodetype == KatNodeType.GROUPED_EXPRESSION:
            value = KatGroupedExpression(arguments=arguments, extra=extra)
        elif nodetype == KatNodeType.EXPRESSION:
            value = KatExpression(
                operator=data["token"], arguments=arguments, extra=extra
            )
        elif nodetype == KatNodeType.ARRAY:
            value = KatArray(arguments=arguments, extra=extra)
        else:
            raise RuntimeError(f"don't know how to generate '{nodetype}'")

        # Detect kwargs.
        if kwarg_extra:
            value = KatKwarg(key=data["key_token"], value=value, extra=kwarg_extra)

        return value


class KatUnparser:
    """TokenContainer to kat script."""

    def unparse(self, container):
        fobj = StringIO()
        self.unparse_file(fobj, container)
        fobj.seek(0)
        return fobj.read()

    def unparse_file(self, fobj, container):
        untokenizer = KatUntokenizer()
        untokenizer.untokenize_file(fobj, container.sorted_tokens)


class KatUntokenizer:
    """Token to kat script."""

    def untokenize(self, tokens):
        fobj = StringIO()
        self.untokenize_file(fobj, tokens)
        fobj.seek(0)
        return fobj.read()

    def untokenize_file(self, fobj, tokens):
        for token in tokens:
            fobj.write(token.raw_value)


class KatSyntaxUnparser:
    """Unparser for kat script meta-language."""

    def __init__(self, spec=None):
        self.unbuilder = KatUnbuilder(spec=spec)

    def syntax(self, directive, optional_as_positional=False):
        """Get syntax for `directive`.

        Parameters
        ----------
        directive : str
            The directive to generate syntax for.

        optional_as_positional : bool, optional
            Show optional parameters in positional form rather than keyword form.
            Defaults to False.

        Returns
        -------
        str
            The syntax for `directive`.

        Raises
        ------
        ValueError
            If `directive` is not recognised as a valid KatScript directive in `spec`.
        """
        from finesse.script.adapter import (
            ElementAdapter,
            CommandAdapter,
            AnalysisAdapter,
        )

        try:
            adapter = self.spec.directives[directive]
        except KeyError as e:
            raise ValueError(f"directive '{directive}' not recognised") from e

        args, has_var_positional, has_var_keyword = self._arguments(
            adapter, optional_as_positional
        )

        if isinstance(adapter, ElementAdapter):
            if has_var_positional:
                args.append("{arg1{ arg2{ ...}}}")
            if has_var_keyword:
                args.append("{kwarg1=value{ kwarg2=value{ ...}}}")

            return f"{directive} {' '.join(arg for arg in args)}"
        elif isinstance(adapter, (CommandAdapter, AnalysisAdapter)):
            if has_var_positional:
                args.append("{arg1{, arg2{, ...}}}")
            if has_var_keyword:
                args.append("{kwarg1=value{, kwarg2=value{, ...}}}")

            return f"{directive}({', '.join(arg for arg in args)})"
        else:
            raise ValueError(f"unrecognised adapter type '{type(adapter)}'")

    def _arguments(self, adapter, optional_as_positional):
        from inspect import Parameter as SigParameter

        args = []
        has_var_positional = False
        has_var_keyword = False

        for arg, parameter in adapter.call_signature().parameters.items():
            if parameter.kind == SigParameter.VAR_POSITIONAL:
                # This is the *args variable argument.
                has_var_positional = True
            elif parameter.kind == SigParameter.VAR_KEYWORD:
                # This is the **kwargs variable argument.
                has_var_keyword = True
            else:
                # Generate real kat script for the default value.
                pos_value = arg

                if parameter.default is parameter.empty:
                    # The parameter does not have a default.
                    key_value = arg
                else:
                    default = self.unbuilder.unbuild(parameter.default)
                    key_value = f"{arg}={default}"

                if parameter.kind == SigParameter.POSITIONAL_ONLY:
                    args.append(pos_value)
                elif parameter.kind == SigParameter.KEYWORD_ONLY:
                    args.append(key_value)
                else:
                    if optional_as_positional:
                        args.append(pos_value)
                    else:
                        args.append(key_value)

        return args, has_var_positional, has_var_keyword

    @property
    def spec(self):
        return self.unbuilder.spec



from finesse.components import Connector, NodeType, NodeDirection
from finesse.components.general import DOFDefinition
from finesse.components.mechanical import MIMOTFWorkspace, mimo_fill, get_mechanical_port
from finesse.components.workspace import ConnectorWorkspace
import importlib.resources as pkg_resources
import numpy as np
from finesse.components import NodeType

TRIPLE_DATA = (
    None  # Global for storing the triple pendulum data so we don't keep reloading it
)

QUAD_DATA = (
    None  # Global for storing the triple pendulum data so we don't keep reloading it
)

def node_2_ligo_node(io, node, name_map={}):
    """Maps a Finesse node to a LIGO suspension model
    degree of freedom string, i.e. in.gnd.disp.L
    """
    if node.name.startswith("F_"):
        _type = "drive"
        dof = node.name[2:]
    else:
        _type = "disp"
        dof = node.name
    mapp = {"z": "L", "yaw": "Y", "pitch": "P"}
    return "{io}.{port}.{_type}.{dof}".format(
        io=io,
        port=name_map.get(node.port.name, node.port.name),
        _type=_type,
        dof=mapp[dof.lower()],
    )


class LIGOQuadSuspension(Connector):
    """A mechanics element that represents a LIGO Quad suspension.

    LIGO Quad
    generate_Quad_Model_Production.m
    SVN REV 10312
    https://redoubt.ligo-wa.caltech.edu/svn/sus/trunk/QUAD/Common/MatlabTools

    The Matlab code above was run to generate the state space model which
    was exported to Python. The Python Controls toolbox was then used
    to generate the transfer-function coefficients and stored in a Pickled
    file which this object accesses.

    The component being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.

    This mechanics element provides access to the ground (`gnd` port)
    and penultimate (`pum` port) mass stages for injecting in
    displacement noise or feedback signals for controlling the test
    (`tst` port) mass.

    >>> from scipy.io import loadmat
    >>> import bz2
    >>> import pickle
    >>> 
    >>> SS = loadmat('/Users/ddb/quad_ss.mat')
    >>> 
    >>> A = SS['ss'][0][0][0]
    >>> B = SS['ss'][0][0][1]
    >>> C = SS['ss'][0][0][2]
    >>> D = SS['ss'][0][0][3]
    >>> inputs = {_[0]: i for i, _ in enumerate(SS['ss'][0][0][15].squeeze())}
    >>> outputs = {_[0]: i for i, _ in enumerate(SS['ss'][0][0][18].squeeze())}
    >>> 
    >>> tfs = control.ss2tf(A, B, C, D)
    >>> 
    >>> with bz2.BZ2File("/Users/ddb/git/finesse3/src/finesse/components/ligo/data/quad_damped_ss.pbz2", 'w') as f:
    >>>     pickle.dump({
    >>>         "A": A,
    >>>         "B": B,
    >>>         "C": C,
    >>>         "D": D,
    >>>         "inputs": inputs,
    >>>         "outputs": outputs
    >>>     }, f, protocol=3)
    """

    def __init__(self, name, connect_to):
        super().__init__(name)
        mech_port = get_mechanical_port(connect_to)
        # Add motion and force nodes to mech port.
        # Here we duplicate the already created mechanical
        # nodes in some other connector element
        self._add_port("tst", NodeType.MECHANICAL)
        self.tst._add_node("z", None, mech_port.z)
        self.tst._add_node("yaw", None, mech_port.yaw)
        self.tst._add_node("pitch", None, mech_port.pitch)
        self.tst._add_node("F_z", None, mech_port.F_z)
        self.tst._add_node("F_yaw", None, mech_port.F_yaw)
        self.tst._add_node("F_pitch", None, mech_port.F_pitch)
        # Suspension point ground port
        self._add_port("gnd", NodeType.MECHANICAL)
        self.gnd._add_node("z", NodeDirection.OUTPUT)
        self.gnd._add_node("yaw", NodeDirection.OUTPUT)
        self.gnd._add_node("pitch", NodeDirection.OUTPUT)
        # Penultimate mass port
        self._add_port("pum", NodeType.MECHANICAL)
        self.pum._add_node("F_z", NodeDirection.INPUT)
        self.pum._add_node("F_yaw", NodeDirection.INPUT)
        self.pum._add_node("F_pitch", NodeDirection.INPUT)
        # Intermediate mass port
        self._add_port("uim", NodeType.MECHANICAL)
        self.uim._add_node("F_z", NodeDirection.INPUT)
        self.uim._add_node("F_yaw", NodeDirection.INPUT)
        self.uim._add_node("F_pitch", NodeDirection.INPUT)

        global QUAD_DATA
        if QUAD_DATA is None:
            import bz2
            import control
            import pickle
            from . import data

            # Load some data in and process it
            ss = pickle.loads(
                bz2.decompress(
                    pkg_resources.read_binary(data, "quad_damped_ss.pbz2")
                )
            )
            tfs = control.ss2tf(ss["A"], ss["B"], ss["C"], ss["D"],)
            QUAD_DATA = (tfs, ss["inputs"], ss["outputs"], control.ss(ss["A"], ss["B"], ss["C"], ss["D"]))

        self.tfs, self.inputs, self.outputs, self._ss = QUAD_DATA

        # Add in connections for GND/UIM/PUM coupling into TST
        for i in self.gnd.nodes + self.pum.nodes + self.uim.nodes:
            for o in self.tst.nodes:
                # Sus model computes how PUM/GND couple
                # into TST displacement
                if not o.name.startswith("F_"):
                    self._register_node_coupling(
                        f"{i.full_name}__{o.full_name}".replace(".", "_"), i, o
                    )
        # Add in TST drives to TST displacements
        # coupling and cross-coupling
        for i in self.tst.nodes:
            if i.name.startswith("F_"):  # drives only
                for o in self.tst.nodes:
                    if not o.name.startswith("F_"):  # disp only
                        self._register_node_coupling(
                            f"{i.full_name}__{o.full_name}".replace(".", "_"), i, o
                        )

        # Define typical degrees of freedom for this component
        import types
        self.dofs = types.SimpleNamespace()
        self.dofs.tst_z = DOFDefinition(None, self.tst.z, 1)
        self.dofs.tst_F_z = DOFDefinition(None, self.tst.F_z, 1)
        self.dofs.pum_F_z = DOFDefinition(None, self.pum.F_z, 1)
        self.dofs.gnd_z = DOFDefinition(None, self.gnd.z, 1)
        
    @property
    def ss(self):
        return self._ss

    def input_output_indices(self, input_node, output_node):
        name_map = {n.port.name: "tst" for n in self.tst.nodes}
        idx = self.inputs[node_2_ligo_node("in", input_node, name_map)]
        odx = self.outputs[node_2_ligo_node("out", output_node, name_map)]
        return idx, odx

    def bode_plot(self, input_node, output_node, omega=None, Hz=True, **kwargs):
        import control
        i,o = self.input_output_indices(input_node, output_node)
        # Reduce numberof inputs/outputs in SS
        A = self.ss.A
        B = self.ss.B[:,i]
        C = self.ss.C[o,:]
        D = self.ss.D[o,i]
        ss = control.ss(A,B,C,D)
        return control.bode_plot(ss, omega=omega, Hz=Hz, **kwargs)

    def _get_workspace(self, sim):
        if sim.signal:
            import control
            refill = sim.model.fsig.f.is_changing  # Need to recompute H(f)
            idxs = []
            odxs = []
            N = 0
            for j, (i, o) in enumerate(self._registered_connections.values()):
                if i in sim.signal.nodes and o in sim.signal.nodes:
                    i = self.nodes[i]
                    o = self.nodes[o]
                    idx, odx = self.input_output_indices(i, o)
                    idxs.append(idx)
                    odxs.append(odx)
                    N += 1

            if N > 0:
                ws = MIMOSSWorkspace(self, sim, refill, refill)
                ws.signal.add_fill_function(self._signal_fill)
                ws.N = N
                ws.iselect = tuple(set(idxs))
                ws.oselect = tuple(set(odxs))
                A = self.ss.A
                B = self.ss.B[:, ws.iselect]
                C = self.ss.C[ws.oselect, :]
                D = self.ss.D[np.ix_(ws.oselect, ws.iselect)]
                ws.idxs = tuple((ws.iselect.index(_) for _ in idxs))
                ws.odxs = tuple((ws.oselect.index(_) for _ in odxs))
                ws.ss = control.ss(
                    self.ss.A,
                    self.ss.B[:, ws.iselect],
                    self.ss.C[ws.oselect, :],
                    self.ss.D[np.ix_(ws.oselect, ws.iselect)]
                )
                return ws
            else:
                return None
        else:
            return None

    def _signal_fill(self, ws):
        """
        There might be smarter more efficient ways to calcuate this:
            A note on shifted Hessenberg systems and frequency response computation
                https://dl.acm.org/doi/10.1145/2049673.2049676
                In this article, we propose a numerical algorithm for efficient
                and robust solution of a sequence of shifted Hessenberg linear systems.
                In particular, we show how the frequency response 𝒢(σ) = d-C(A-σ I)-1b in
                the single input case can be computed more efficiently than with other
                state-of-the-art methods. We also provide a backward stability analysis
                of the proposed algorithm. 
        """
        s = 2j * np.pi * ws.sim.model_data.fsig
        H = ws.ss(s)
        for i in range(ws.N):
            with ws.sim.signal.component_edge_fill3(ws.owner_id, i, 0, 0) as mat:
                mat[:] = H[ws.odxs[i], ws.idxs[i]]


class MIMOSSWorkspace(ConnectorWorkspace):
    pass


class LIGOTripleSuspension(Connector):
    """A mechanics element that represents a LIGO Triple suspension.

    LIGO Triple version 20140304TMproductionTM
    generate_TRIPLE_Model_Production.m
    SVN REV 10312
    https://redoubt.ligo-wa.caltech.edu/svn/sus/trunk/QUAD/Common/MatlabTools

    The Matlab code above was run to generate the state space model which
    was exported to Python. The Python Controls toolbox was then used
    to generate the transfer-function coefficients and stored in a Pickled
    file which this object accesses.

    The component being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.

    This mechanics element provides access to the ground (`gnd` port)
    and penultimate (`pum` port) mass stages for injecting in
    displacement noise or feedback signals for controlling the test
    (`tst` port) mass.
    """

    def __init__(self, name, connect_to):
        super().__init__(name)
        mech_port = get_mechanical_port(connect_to)
        # Add motion and force nodes to mech port.
        # Here we duplicate the already created mechanical
        # nodes in some other connector element
        self._add_port("tst", NodeType.MECHANICAL)
        self.tst._add_node("z", None, mech_port.z)
        self.tst._add_node("yaw", None, mech_port.yaw)
        self.tst._add_node("pitch", None, mech_port.pitch)
        self.tst._add_node("F_z", None, mech_port.F_z)
        self.tst._add_node("F_yaw", None, mech_port.F_yaw)
        self.tst._add_node("F_pitch", None, mech_port.F_pitch)
        # Suspension point ground port
        self._add_port("gnd", NodeType.MECHANICAL)
        self.gnd._add_node("z", NodeDirection.OUTPUT)
        self.gnd._add_node("yaw", NodeDirection.OUTPUT)
        self.gnd._add_node("pitch", NodeDirection.OUTPUT)
        # Penultimate mass port
        self._add_port("pum", NodeType.MECHANICAL)
        self.pum._add_node("F_z", NodeDirection.INPUT)
        self.pum._add_node("F_yaw", NodeDirection.INPUT)
        self.pum._add_node("F_pitch", NodeDirection.INPUT)

        global TRIPLE_DATA
        if TRIPLE_DATA is None:
            import bz2
            import control
            import pickle
            from . import data

            # Load some data in and process it
            ss = pickle.loads(
                bz2.decompress(
                    pkg_resources.read_binary(data, "ligo_triple_suspension_ss.pbz2")
                )
            )
            tfs = control.ss2tf(ss["A"], ss["B"], ss["C"], ss["D"],)
            TRIPLE_DATA = (tfs, ss["inputs"], ss["outputs"])

        self.tfs, self.inputs, self.outputs = TRIPLE_DATA

        # Add in connections for GND/PUM coupling into TST
        for i in self.gnd.nodes + self.pum.nodes:
            for o in self.tst.nodes:
                # Sus model computes how PUM/GND couple
                # into TST displacement
                if not o.name.startswith("F_"):
                    self._register_node_coupling(
                        f"{i.full_name}__{o.full_name}".replace(".", "_"), i, o
                    )
        # Add in TST drives to TST displacements
        # coupling and cross-coupling
        for i in self.tst.nodes:
            if i.name.startswith("F_"):  # drives only
                for o in self.tst.nodes:
                    if not o.name.startswith("F_"):  # disp only
                        self._register_node_coupling(
                            f"{i.full_name}__{o.full_name}".replace(".", "_"), i, o
                        )

        # Define typical degrees of freedom for this component
        import types
        self.dofs = types.SimpleNamespace()
        self.dofs.tst_z = DOFDefinition(None, self.tst.z, 1)
        self.dofs.tst_F_z = DOFDefinition(None, self.tst.F_z, 1)
        self.dofs.pum_F_z = DOFDefinition(None, self.pum.F_z, 1)
        self.dofs.gnd_z = DOFDefinition(None, self.gnd.z, 1)
        
    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing  # Need to recompute H(f)
            N = len(self._registered_connections)
            ws = MIMOTFWorkspace(self, sim, refill, N)
            ws.set_denominator(self.tfs.den[0][0])
            name_map = {n.port.name: "tst" for n in self.tst.nodes}

            # Setup the TFs for filling
            for j, (i, o) in enumerate(self._registered_connections.values()):
                i = self.nodes[i]
                o = self.nodes[o]
                idx = self.inputs[node_2_ligo_node("in", i, name_map)]
                odx = self.outputs[node_2_ligo_node("out", o, name_map)]
                ws.add_numerator(self.tfs.num[odx][idx])
                
            return ws
        else:
            return None



"""Tests for u_nm function correctness"""

import numpy as np
from scipy.special import factorial, hermite
from numpy.testing import assert_allclose
import pytest
from finesse.gaussian import HGMode


def analytic_un(n, x, qx):
    """Direct translation of Equation (9.34) from:
    'Interferometer Techniques for Gravitational Wave Detection'
    """
    k = 2 * np.pi / qx.wavelength
    return (
        (2 / np.pi) ** 0.25
        * np.sqrt(1 / (2 ** n * factorial(n) * qx.w0))
        * np.sqrt(1j * qx.zr / qx.q)
        * (1j * qx.zr * qx.conjugate().q / (-1j * qx.zr * qx.q)) ** (n / 2)
        * hermite(n)(np.sqrt(2) * x / qx.w)
        * np.exp(-1j * k * x ** 2 / (2 * qx.q))
    )


@pytest.mark.parametrize(
    "qx_qy", ([1j, 1j], [-1 + 0.6j, -1.1 + 0.6j], [3 + 1.5j, 4.2 + 1.4j]),
)
@pytest.mark.parametrize(
    "n", (0, 1, 4, 9, 22),
)
@pytest.mark.parametrize(
    "m", (0, 2, 3, 6, 23),
)
def test_unm(qx_qy, n, m):
    """Test that u_nm functions in finesse.cymath.homs give correct
    results corresponding to exact analytic form."""
    qx, qy = qx_qy
    HGnm = HGMode([qx, qy], n, m)

    x = np.linspace(-HGnm.qx.w0 * 3, HGnm.qx.w0 * 3, 51)
    y = np.linspace(-HGnm.qy.w0 * 3, HGnm.qy.w0 * 3, 51)
    u_nm = HGnm.unm(x, y)

    expect = np.outer(analytic_un(n, x, HGnm.qx), analytic_un(m, y, HGnm.qy))

    assert_allclose(u_nm, expect)



"""Model copy unit test fixtures."""

import pytest
from finesse import Model
from finesse.components import Laser, Beamsplitter, Mirror, Cavity, Gauss
from finesse.detectors import PowerDetector
from finesse.analysis import xaxis


# FIXME: this fixture should provide a very complicated model, ideally one with every type of
# component.
@pytest.fixture(scope="module")
def complicated_ifo_for_copying():
    """Fixture for constructing a complicated model for testing copy operations."""
    model = Model()

    model.add(
        [
            Laser("L0"),
            Beamsplitter("BS"),
            Mirror("IMX", R=0.99, L=0, Rc=-10),
            Mirror("IMY", R=0.99, L=0, Rc=-10),
            Mirror("EMX", T=10e-6, L=0, Rc=10),
            Mirror("EMY", T=10e-6, L=0, Rc=10),
        ]
    )
    model.connect(model.L0.p1, model.BS.p1)
    model.connect(model.BS.p2, model.IMY.p1, L=1)
    model.connect(model.IMY.p2, model.EMY.p1, L=1)
    model.connect(model.BS.p3, model.IMX.p1, L=1)
    model.connect(model.IMX.p2, model.EMX.p1, L=1)

    model.add(PowerDetector("refl", model.IMX.p1.o))
    model.add(PowerDetector("circ", model.EMX.p1.i))
    model.add(PowerDetector("trns", model.EMX.p2.o))

    model.add(Cavity("cav_y", model.IMY.p2.o, model.IMY.p2.i))
    model.add(Cavity("cav_x", model.IMX.p2.o, model.IMX.p2.i))
    model.add(Gauss("g_L0", model.L0.p1.o, q=(-0.2 + 2.5j)))

    # model.select_modes("even", 4)

    xaxis(model.EMX.phi, "lin", -180, 180, 400)

    return model



"""Finesse datastore tools.

The datastore is intended for objects that should be cached during the execution of the
current Python kernel. Code typically uses this instead of the more bug-prone singleton
pattern (see #260).
"""

_DATASTORE = {}


def invalidate(key=None):
    """Invalidate the datastore.

    Parameters
    ----------
    key : str, optional
        The datastore key to invalidate.
    """
    if key is None:
        _DATASTORE.clear()
    else:
        del _DATASTORE[key]


def init_singleton(cls, *args, **kwargs):
    """Instantiate `cls` and return the object for the current and future calls.

    Parameters
    ----------
    cls : type
        The singeton class to retrieve. If `cls` has already been instantiated, the
        existing instance is returned and `args` and `kwargs` are ignored.

    Other Parameters
    ----------------
    args, kwargs
        Positional and keyword arguments to pass to the `cls` call, if `cls` is not yet
        instantiated.

    Returns
    -------
    object
        The instantiated singleton.
    """
    if cls not in _DATASTORE:
        _DATASTORE[cls] = cls(*args, **kwargs)
    return _DATASTORE[cls]


def has_singleton(cls):
    """Determine if `cls` is instantiated.

    Parameters
    ----------
    cls : type
        The singeton class to check.

    Returns
    -------
    bool
        `True` if an instance exists, `False` otherwise.
    """
    return cls in _DATASTORE


def invalidate_singleton(cls):
    """Invalidate a singleton object with type `cls`.

    Parameters
    ----------
    cls : type
        The singeton class to invalidate.
    """
    invalidate(cls)



"""HOM quantum noise tests."""

import pytest
import numpy as np
from finesse.analysis import x2axis
from scipy import constants
from numpy.testing import assert_allclose


@pytest.fixture
def lossy_mirror_model(model):
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 m1.p1 L=0
        m m1 R=0.5 T=0.4

        qnoised qd_laser l1.p1.o
        qnoised qd_trans m1.p2.o
        qnoised qd_refl m1.p1.o

        fsig(1)
    """
    )

    return model


@pytest.fixture(
    params=(
        ("m comp R=0.5 T=0.4", "p2"),
        ("lens comp f=1", "p2"),
        ("bs comp R=0.4 T=0.3", "p2"),
        ("bs comp R=0.4 T=0.3", "p3"),
        ("mod comp mod_type=am midx=0.4 f=10M", "p2"),
        ("mod comp mod_type=pm midx=0.4 f=10M order=3", "p2"),
        ("dbs comp", "p3"),
    ),
    ids=(
        "mirror",
        "lens",
        "beamsplitter_reflection",
        "beamsplitter_transmission",
        "amplitude_modulator",
        "phase_modulator",
        "directional_beamsplitter",
    ),
)
def mismatch_model(model, request):
    element = request.param[0]
    port = request.param[1]
    model.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 comp.p1 L=0
        {element}

        qnoised qd_laser l1.p1.o
        qnoised qd_out comp.{port}.o

        fsig(1)
        modes(modes="even", maxtem=0)

        var w0 1
        var z 1

        gauss g1 comp.p1.o w0=&w0 z=&z
        gauss g2 comp.{port}.o w0=1 z=1

        pd pd_out comp.{port}.o
    """
    )

    return model


@pytest.fixture
def modulator_model(model):
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 mod1.p1 L=0
        mod mod1 f=10M midx=0.8 mod_type=am

        qnoised qd1 mod1.p2.o
        qnoised1 qd2 mod1.p2.o &mod1.f 0

        pd pd1 mod1.p2.o
        pd1 pd2 mod1.p2.o &mod1.f 0

        fsig(1)
        modes(maxtem=2)
        gauss g1 l1.p1.o w0=1 z=0
    """
    )

    return model


@pytest.fixture
def squeezer_model(model):
    model.parse(
        """
        sq sqz 10
        s s1 sqz.p1 mod.p1 L=5
        mod mod f=10M midx=0.5
        s s2 mod.p2 m1.p1 L=5
        m m1 R=0.9 T=0.1
        s scav m1.p2 m2.p1 L=30
        m m2 R=0.9 T=0.1

        free_mass fm m2.mech mass=1

        qnoised qd1 m1.p1.o
        qnoised1 qd2 m1.p1.o &mod.f 0
        qnoised1 qd3 m1.p1.o -&mod.f 0
        qnoised1 qd4 m1.p1.o &mod.f 0
        qnoised1 qd5 m1.p1.o -&mod.f 0

        fsig(1)
        modes(maxtem=2)
        gauss g1 sqz1.p1.o w0=1 z=0
    """
    )

    return model


def test_vacuum_quantum_noise(lossy_mirror_model):
    """Test basic quantum noise generation with homs."""
    lossy_mirror_model.parse(
        """
        modes(modes="even", maxtem=2)
        gauss g2 m1.p2.o w0=1 z=1

        xaxis(l1.f, log, 1, 1e9, 100)  # Sweep laser frequency from 1Hz to 1GHz
        """
    )

    out = lossy_mirror_model.run()
    xaxis = out.x[0]

    # Should just see Schottky noise, 2Phf
    qn = (
        2
        * lossy_mirror_model.l1.P.value
        * constants.h
        * (lossy_mirror_model.f0 + xaxis)
    )
    qn_asd = np.sqrt(qn)
    qn_asd_trans = qn_asd * np.sqrt(lossy_mirror_model.m1.T.value)
    qn_asd_refl = qn_asd * np.sqrt(lossy_mirror_model.m1.R.value)

    assert_allclose(out["qd_laser"], qn_asd, rtol=1e-14, atol=0)
    assert_allclose(out["qd_trans"], qn_asd_trans, rtol=1e-14, atol=0)
    assert_allclose(out["qd_refl"], qn_asd_refl, rtol=1e-14, atol=0)


@pytest.mark.xfail(
    reason=(
        "Waist size/position mismatches with maxtem > 0 are incorrect in both Finesse 2 and 3, "
        "so marking these as expected failures for now"
    )
)
@pytest.mark.parametrize("maxtem", (0, 2, 4))
@pytest.mark.parametrize("axis", ("w0", "z"))
def test_vacuum_quantum_noise_mismatch(mismatch_model, axis, maxtem):
    """Test quantum noise generation with a mismatch across various components."""
    model = mismatch_model
    model.select_modes(maxtem=maxtem)

    out = x2axis(
        model.l1.f, "lin", -3e6, 3e6, 6, getattr(model, axis).value, "lin", 0.5, 2, 5
    )

    # Should just see Schottky noise, 2Phf
    qn = 2 * model.l1.P.value * constants.h * (model.f0 + out.x[0])
    qn_asd = np.sqrt(qn)

    # Tile the output up to the correct shape
    qn_asd = np.tile(qn_asd[:, None], (1, len(out.x[1])))

    assert_allclose(out["qd_laser"], qn_asd, rtol=1e-14, atol=0)
    assert_allclose(out["qd_out"] / np.sqrt(out["pd_out"]), qn_asd, rtol=1e-14, atol=0)


def test_modulator_quantum_noise_sweep_laser_freq(modulator_model):
    """Test modulator quantum noise couplings with varying laser frequency."""
    modulator_model.parse(
        """
        xaxis(l1.f, log, 1, 1e9, 100)  # Sweep laser frequency from 1Hz to 1GHz
        """
    )

    out = modulator_model.run()

    # Should just see Schottky noise, 2Phf
    qn = 2 * constants.h * (modulator_model.f0 + out.x[0])
    qn_asd = np.sqrt(qn)

    assert_allclose(out["qd1"] / np.sqrt(out["pd1"]), qn_asd, rtol=1e-14, atol=0)

    # TODO: work out the correct analytics for this
    # assert_allclose(out["qd2"] / np.sqrt(out["pd2"]), qn_asd, rtol=1e-14, atol=0)


def test_modulator_quantum_noise_sweep_modulation_index(modulator_model):
    """Test modulator quantum noise couplings with varying modulation index."""
    modulator_model.parse(
        """
        xaxis(mod1.midx, lin, 0, 1, 100)
        """
    )

    out = modulator_model.run()

    # Should just see Schottky noise, 2Phf
    qn = 2 * constants.h * modulator_model.f0
    qn_asd = np.sqrt(qn)

    assert_allclose(out["qd1"] / np.sqrt(out["pd1"]), qn_asd, rtol=1e-14, atol=0)


@pytest.mark.skip(reason="Not valid until finesse 3 handles this properly")
def test_modulator_quantum_noise_squeezer(squeezer_model):
    """Test modulator quantum noise couplings with squeezed light input."""
    squeezer_model.parse(
        """
        xaxis(sqz.f, lin, -10, 10, 20)  # Sweep squeezer frequency from -10Hz to +10Hz
        """
    )

    out = squeezer_model.run()
    xaxis = out.x[0]
    qn = 0.5 * (1 + xaxis / squeezer_model.f0) * constants.h

    assert_allclose(out["qd1"], qn, rtol=1e-5, atol=1e-8)
    assert_allclose(out["qd2"], qn, rtol=1e-5, atol=1e-8)
    assert_allclose(out["qd3"], qn, rtol=1e-5, atol=1e-8)



"""KatScript lexer for syntax highlighting with Pygments."""


from pygments.lexer import Lexer, RegexLexer, using, bygroups
import pygments
import pygments.token as token
import re


class KatScriptPygmentsLexer(Lexer):
    name = "KatScript"
    aliases = ["katscript", "kat"]
    filenames = ["*.kat"]

    # Default map of KatToken types to pygments.token types
    token_map = {
        "COMMENT": token.Comment,
        "BOOLEAN": token.Keyword.Reserved,
        "REFERENCE": token.Name.Variable,
        "NUMBER": token.Number,
        "DIVIDE": token.Operator,
        "EQUALS": token.Operator,
        "MINUS": token.Operator,
        "PLUS": token.Operator,
        "TIMES": token.Operator,
        "COMMA": token.Punctuation,
        "LPAREN": token.Punctuation,
        "RPAREN": token.Punctuation,
        "LBRACKET": token.Punctuation,
        "RBRACKET": token.Punctuation,
        "STRING": token.String,
        "NAME": token.Text,
        "NEWLINE": token.Text,
        "WHITESPACE": token.Text,
    }

    def get_tokens_unprocessed(self, text):
        from finesse.script.parser import KatParser
        import finesse.script.containers as containers

        parser = KatParser()
        script = parser.parse(text)

        # Many tokens are typed as "NAME", with no other information available in
        # `script.sorted_tokens`. We therefore first do a pass through `script.arguments` (where
        # this information is available), and tag those that have some special meaning.
        for arg in script.arguments:
            if type(arg) == containers.KatFunction:
                arg.directive._syntax_highlight_type = token.Name.Function.Magic
            elif type(arg) == containers.KatElement:
                arg.directive._syntax_highlight_type = token.Keyword.Type
                arg.name._syntax_highlight_type = token.Text

        for tok in script.sorted_tokens:
            if type(tok) == containers.KatMetaToken:
                continue
            elif hasattr(tok, "_syntax_highlight_type"):
                # If we've already tagged this token, just use the tagged type
                yield tok.start_index, tok._syntax_highlight_type, tok.raw_value
            elif tok.type == "NAME" and "." in tok.raw_value:
                # Assume any "foo.bar" style tokens are variables
                split = tok.raw_value.split(".")
                yield tok.start_index, token.Text, split[0]

                pos = tok.start_index + len(split[0])
                for s in split[1:]:
                    yield pos, token.Punctuation, "."
                    pos += 1
                    yield pos, token.Text, s
                    pos += len(s)
            elif tok.type == "REFERENCE":
                # Don't highlight the initial "&" in a reference, as it's really an operator
                yield tok.start_index, token.Operator, tok.raw_value[0]
                yield tok.start_index + 1, self.token_map[tok.type], tok.raw_value[1:]
            else:
                yield tok.start_index, self.token_map[tok.type], tok.raw_value


class KatScriptSubstringPygmentsLexer(RegexLexer):
    name = "KatScriptInPython"
    aliases = ["katscriptinpython"]

    flags = re.DOTALL

    tokens = {
        "root": [
            (
                r'(.+?""")([\s\n]*#kat)',
                bygroups(using(pygments.lexers.python.PythonLexer), None),
                "katscript",
            ),
            (".+", using(pygments.lexers.python.PythonLexer)),
        ],
        "katscript": [
            (
                '(.*?)(""")',
                bygroups(
                    using(KatScriptPygmentsLexer),
                    using(pygments.lexers.python.PythonLexer),
                ),
                "#pop",
            ),
        ],
    }
